/*
 * $Id: FlashFile.java,v 1.11 2001/07/25 22:44:45 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api;

import java.io.*;
import java.util.*;
import com.iv.flash.util.*;
import com.iv.flash.url.*;
import com.iv.flash.log.*;
import com.iv.flash.cache.*;
import com.iv.flash.parser.*;
import com.iv.flash.api.text.*;

/**
 * Class represents a Flash file (either .swt or .swf)
 *
 * @author Dmitry Skavish
 */
public final class FlashFile {

    private String fileName;                // file name
    private String fileDir;                 // file directory

    private int version;                    // version of flash file format
    private Rect frameSize;                 // size of the movie in twixels
    private int frameRate;                  // frame rate (number of frames/second left shifted by 8)
    private int fileSize;                   // file size in bytes
    private Script main;                    // main timeline

    private int processCount = 0;
    private boolean isTemplate = false;     // true - swt, false - swf
    private FlashFile defaultSymbolFile;    // file of default flash symbols

    private IVMap defsByID = new IVMap();   // all definitions defined in the file

    private Hashtable defsByName = new Hashtable();     // all definitions defined and named in the file

    private Hashtable externalFiles = new Hashtable();  // all external media loaded from within the file

    public FlashFile() {}

    public String getFileName() {
        return fileName;
    }
    public void setFileName( String fileName ) {
        this.fileName = Util.translatePath(fileName);
    }

    public String getFileDir() {
        return fileDir;
    }
    public void setFileDir( String fileDir ) {
        this.fileDir = Util.translatePath(fileDir);
    }

    /**
     * Returns absolute file name
     *
     * @return absolute file name
     */
    public String getFullName() {
        return new File( fileDir, fileName ).getAbsolutePath();
    }

    public void setVersion( int version ) {
        this.version = version;
    }

    public void setFrameSize( Rect r ) {
        this.frameSize = r;
    }

    public void setFileSize( int size ) {
        this.fileSize = size;
    }

    public void setFrameRate( int fr ) {
        this.frameRate = fr;
    }

    public void setMainScript( Script main ) {
        this.main = main;
    }

    public int getVersion() {
        return version;
    }

    public Rect getFrameSize() {
        return frameSize;
    }

    public int getFileSize() {
        return fileSize;
    }

    public int getFrameRate() {
        return frameRate;
    }

    public Script getMainScript() {
        return main;
    }

    public void setTemplate( boolean isTemplate ) {
        this.isTemplate = isTemplate;
    }

    /**
     * Returns true if the file is Flash template, otherwise returns false
     * <p>
     * The criteria is not the file extention (.swt), but MM Generator special 'template' tag
     */
    public boolean isTemplate() {
        return isTemplate;
    }

    /**
     * Returns definition by ID
     *
     * @param id     definition ID
     * @return definition or null
     */
    public FlashDef getDef( int id ) {
        return defsByID.get(id);
    }

    /**
     * Adds definition by ID
     *
     * @param def    definition to be added
     */
    public void addDef( FlashDef def ) {
        defsByID.add(def);
    }

    /**
     * Adds definition by name
     *
     * @param name   name of the definition
     * @param def    definition
     */
    public void addDefToLibrary( String name, FlashDef def ) {
        if( !PropertyManager.symCaseSensitive ) name = name.toUpperCase();
        defsByName.put(name, def);
    }

    /**
     * Finds definition by name in this file and all its external files (if any)
     *
     * @param name   name of the definition
     * @return found definition or null
     */
    public FlashDef getDefFromLibrary( String name ) {
        if( !PropertyManager.symCaseSensitive ) name = name.toUpperCase();
        FlashDef def = (FlashDef) defsByName.get(name);
        if( def != null ) return def;
        if( externalFiles.size() > 0 ) {
            Enumeration files = externalFiles.elements();
            while( files.hasMoreElements() ) {
                FlashFile file = (FlashFile) files.nextElement();
                def = file.getDefFromLibrary( name );
                if( def != null ) return def;
            }
        }
        return null;
    }

    /**
     * Reads, parses, and adds external file to this file (if it is not here yet)
     *
     * @param fileName name of the file to be added
     * @param cache    if true then cache the file in the MediaCache
     * @return added flash file
     * @exception IVException
     */
    public synchronized FlashFile addExternalFile( String fileName, boolean cache ) throws IVException {
        try {
            IVUrl url = IVUrl.newUrl(fileName, this);
            fileName = url.getName();
            FlashFile file = getExternalFile( fileName );
            if( file != null ) return file;
            // try to retrive from media cache
            file = (FlashFile) MediaCache.getMedia( fileName );
            if( file == null ) {
                InputStream is = url.getInputStream();
                file = FlashFile.parse(fileName, is);
                MediaCache.addMedia( url, file, file.getFileSize(), cache );
            }
            addExternalFile( fileName, file );
            return file;
        } catch( IOException e ) {
            throw new IVException( e, "errReadingFile", fileName );
        }
    }

    /**
     * Finds external file by file name
     *
     * @param filename name of the file to be searched
     * @return found file or null
     */
    public FlashFile getExternalFile( String filename ) {
        return (FlashFile) externalFiles.get( filename );
    }

    /**
     * Adds external file to this file (if it is not here yet)
     *
     * @param filename name of the file to be added
     * @param file     flash file to be added
     */
    public void addExternalFile( String filename, FlashFile file ) {
        externalFiles.put( filename, file );
    }

    /**
     * Returns default symbol file
     *
     * @return default symbol file
     */
    public synchronized FlashFile getDefaultSymbolFile() {
        if( defaultSymbolFile != null ) return defaultSymbolFile;
        String fileName = PropertyManager.getProperty("com.iv.flash.defaultSymbolFile","bin/DefaultSymbolFile.swt");
        fileName = Util.getSysFile(fileName).getAbsolutePath();
        try {
            defaultSymbolFile = addExternalFile( fileName, true );
            return defaultSymbolFile;
        } catch( IVException e ) {
            Log.log(e);
            return null;
        }
    }

    /**
     * Searches font by its name in this file and all its external files
     *
     * @param name   font name
     * @return font or null
     */
    public Font getFont( String name ) {
        for( Enumeration e = defsByID.values(); e.hasMoreElements(); ) {
            FlashDef def = (FlashDef) e.nextElement();
            if( def instanceof FontDef ) {
                Font font = ((FontDef)def).getFont();
                if( font.getFontName().equalsIgnoreCase(name) ) return font;
            }
        }
        if( externalFiles.size() > 0 ) {
            Enumeration files = externalFiles.elements();
            while( files.hasMoreElements() ) {
                FlashFile file = (FlashFile) files.nextElement();
                Font font = file.getFont( name );
                if( font != null ) return font;
            }
        }
        return null;
    }

    /**
     * Returns all fonts defined within this file (fonts from external files are not included)
     *
     * @return vector of FontDef objects
     */
    public IVVector getLocalFonts() {
        IVVector v = new IVVector();
        Enumeration e = definitions();
        while( e.hasMoreElements() ) {
            FlashDef def = (FlashDef) e.nextElement();
            if( def instanceof FontDef ) v.addElement( ((FontDef)def).getFont() );
        }
        return v;
    }

    /**
     * Finds script by name defined within this file or any external files
     *
     * @param name   script name
     * @return found script or null
     */
    public Script getScript( String name ) {
        FlashDef def = getDefFromLibrary( name );
        if( def instanceof Script ) return (Script) def;
        return null;
    }

    /**
     * Returns Enumeration of all definitions (FlashDef) defined within this file
     *
     * @return Enumeration of FlashDef
     */
    public Enumeration definitions() {
        return new Enumeration() {
            private Enumeration e = defsByID.values();
            private int names = 0;
            public boolean hasMoreElements() {
                for(;;) {
                    switch(names) {
                        case 0:
                            if( e.hasMoreElements() ) return true;
                            e = defsByName.elements();
                            names = 1;
                        case 1:
                            if( e.hasMoreElements() ) return true;
                            names = 2;
                        case 2:
                            return false;
                    }
                }
            }
            public Object nextElement() {
                return e.nextElement();
            }
        };
    }

    /**
     * Processes this file in the specified context
     *
     * @param context context to process in
     * @return this file
     * @exception IVException
     */
    public FlashFile processFile( Context context ) throws IVException {
        if( !isTemplate() ) return this;
        processScript( main, new StandardContext( context ) );
        return this;
    }

    /**
     * Processes the specified script in the specified context
     *
     * @param script  specified script
     * @param context specified context
     * @return processed script
     * @exception IVException
     */
    public Script processScript( Script script, Context context ) throws IVException {
        processCount++;
        if( processCount < 50 ) {
            if( context == null ) context = new StandardContext();
            script.process(this, context);
        } else {
            throw new IVException( "infiniteLoop" );
        }
        processCount--;
        script.setProcessed();
        return script;
    }

    /**
     * Generates file into FlashOutput buffer
     *
     * @return output buffer
     * @exception IVException
     */
    public FlashOutput generate() throws IVException {
        FlashOutput fob = new FlashOutput( getFileSize() );

        // write signature
        fob.writeByte( 'F' );
        fob.writeByte( 'W' );
        fob.writeByte( 'S' );
        // write version
        fob.writeByte( version );
        // skip file size
        fob.skip(4);
        // write file rect
        frameSize.write( fob );
        // write frame rate
        fob.writeWord( frameRate );

        // write main timeline
        main.generate( fob );

        // write file size
        fob.writeDWordAt( fob.getSize(), 4 );

        return fob;
    }

    /**
     * Parses Flash file from input stream
     *
     * @param fileName file name, used only for establishing the file's current directory
     * @param is       input stream used to read the file's content
     * @return created flash file
     * @exception IVException
     */
    public static FlashFile parse( String fileName, InputStream is ) throws IVException {
        FlashFile file = newFlashFileForParse( fileName );

        try {
            Parser parser = new Parser();
            parser.parseStream( is, file );
            return file;
        } catch( IOException e ) {
            throw new IVException( e, "errParseTemplate", file.getFullName() );
        } finally {
            try {
                is.close();
            } catch( Exception ee ) {}
        }
    }

    /**
     * Parses Flash file from buffer
     *
     * @param fileName file name, used only for establishing the file's current directory
     * @param fb       flash buffer to read the file from
     * @return created flash file
     * @exception IVException
     */
    public static FlashFile parse( String fileName, FlashBuffer fb ) throws IVException {
        FlashFile file = newFlashFileForParse( fileName );
        Parser parser = new Parser();
        parser.parseBuffer( fb, file );
        return file;
    }

    /**
     * Parses Flash file from file
     *
     * @param fileName file name of the flash file
     * @return created flash file
     * @exception IVException
     * @exception FileNotFoundException
     */
    public static FlashFile parse( String fileName ) throws IVException, FileNotFoundException {
        File file = new File(fileName);
        String path = file.getAbsolutePath();

        try {
            FileInputStream is = new FileInputStream( file );
            FlashFile flashFile = parse( path, is );
            return flashFile;
        } catch( FileNotFoundException e ) {
            throw e;
        } catch( IOException e ) {
            throw new IVException( e, "errParseTemplate", path );
        }
    }

    /**
     * Creates empty flash file with default settings
     *
     * @return default flash file
     */
    public static FlashFile newFlashFile() {
        FlashFile file = new FlashFile();
        file.setVersion( 4 );
        file.setFrameSize( new Rect(0, 0, 540*20, 400*20) );
        file.setFrameRate( 12<<8 );
        file.setFileSize( 100 );    // does not have much sense

        return file;
    }

    /**
     * Creates a copy of this flash file
     *
     * @return copy of this flash file
     */
    public synchronized FlashFile copyFile() {
        FlashFile file = new FlashFile();
        file.setFileName( fileName );
        file.setFileDir( fileDir );
        file.setVersion( version );
        file.setFrameSize( (Rect) frameSize.getCopy(null) );
        file.setFrameRate( frameRate );
        file.setFileSize( fileSize );
        file.processCount = processCount;
        file.isTemplate = isTemplate;

        // copy main script
        ScriptCopier copier = new ScriptCopier();
        file.main = (Script) main.getCopy( copier );

        // copier now contains all copied definitions
        // copy defsByID
        for( Enumeration e = defsByID.values(); e.hasMoreElements(); ) {
            FlashDef def = (FlashDef) e.nextElement();
            FlashDef myDef = copier.get( def );
            if( myDef != null ) file.defsByID.add( myDef );
        }

        // copy defsByName
        for( Enumeration e = defsByName.keys(); e.hasMoreElements(); ) {
            Object key = e.nextElement();
            FlashDef def = (FlashDef) defsByName.get( key );
            FlashDef myDef = copier.get( def );
            if( myDef != null ) file.defsByName.put( key, myDef );
        }

        // it's enough just to clone external files
        file.externalFiles = (Hashtable) externalFiles.clone();

        return file;
    }

    public void printContent( PrintStream out ) {
        out.println( "Flash: version="+version+" size="+getFileSize() );
        out.println( "  frame: ("+(frameSize.getXMin()/20)+","+(frameSize.getYMin()/20)+") "+
                     "("+(frameSize.getXMax()/20)+","+(frameSize.getYMax()/20)+")" );
        out.println( "  frameRate="+(frameRate>>8) );
        main.printContent( out, "" );
    }

    private static FlashFile newFlashFileForParse( String fileName ) {
        fileName = Util.translatePath(fileName);
        int index = fileName.lastIndexOf( Util.fileSeparator );
        String dir, name;
        if( index >= 0 ) {
            dir = fileName.substring(0, index);
            name = fileName.substring(index+1);
        } else {
            dir = "";
            name = fileName;
        }
        FlashFile flashFile = new FlashFile();
        flashFile.setFileName( name );
        flashFile.setFileDir( dir );
        return flashFile;
    }

}

