/*
 * $Id: Matrix.java,v 1.3 2001/07/26 22:56:36 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api;

import com.iv.flash.parser.Parser;
import com.iv.flash.util.*;
import java.io.PrintStream;

/**
 * Transformation matrix
 * <P>
 * It represents a standard 2x3 transformation matrix.
 * It is used to describe the scale, rotation and translation of a graphic object.
 * <P>
 * The scaleX, scaleY, rotateSkew0 and rotateSkew1 fields are stored as 16.16 fixed-point values.
 * The translateX and translateY values are stored as signed values in twips.
 * <p>
 * The MATRIX record is optimized for common cases such as matrix that performs a translation only.
 * In this case the HasScale and HasRotate flags are zero, and the matrix only contains the translateX and
 * translateY fields.
 * <p>
 * The mapping from the MATRIX fields to the 2x3 matrix is as follows:<br>
 * <pre>
 * | ScaleX       RotateSkew0  |
 * | RotateSkew1  ScaleY       |
 * | TranslateX   TranslateY   |
 * </pre>
 * <p>
 * For any coordinates (x, y), the transformed coordinates  (x', y') are calculated as follows:<br>
 * <pre><code>
 * x' = x * ScaleX + y * RotateSkew1 + TranslateX
 * y' = x * RotateSkew0 + y * ScaleY + TranslateY
 * </code></pre>
 *
 * @author Dmitry Skavish
 */
public final class Matrix extends FlashItem {

    private double scaleX;
    private double scaleY;
    private double rotateSkew0;
    private double rotateSkew1;
    private int    translateX;
    private int    translateY;

    /**
     * Creates default matrix
     */
    public Matrix() {
        scaleX      = 1.0;
        scaleY      = 1.0;
        rotateSkew0 = 0.0;
        rotateSkew1 = 0.0;
        translateX  = 0;
        translateY  = 0;
    }

    public void setScaleX( double scaleX ) {
        this.scaleX = scaleX;
    }

    public double getScaleX() {
        return scaleX;
    }

    public void setScaleY( double scaleY ) {
        this.scaleY = scaleY;
    }

    public double getScaleY() {
        return scaleY;
    }

    public void setRotateSkew0( double rotateSkew0 ) {
        this.rotateSkew0 = rotateSkew0;
    }

    public double getRotateSkew0() {
        return rotateSkew0;
    }

    public void setRotateSkew1( double rotateSkew1 ) {
        this.rotateSkew1 = rotateSkew1;
    }

    public double getRotateSkew1() {
        return rotateSkew1;
    }

    public void setTranslateX( int translateX ) {
        this.translateX = translateX;
    }

    public void setTranslateX( double translateX )  {
        this.translateX = (int) translateX;
    }

    public int getTranslateX() {
        return translateX;
    }

    public void setTranslateY( int translateY ) {
        this.translateY = translateY;
    }

    public void setTranslateY( double translateY ) {
        this.translateY = (int) translateY;
    }

    public int getTranslateY() {
        return translateY;
    }

    /**
     * Translates X on the specified value
     *
     * @param x      this value will be added to the current one
     */
    public void translateX( double x ) {
        this.translateX += x;
    }

    /**
     * Translates Y on the specified value
     *
     * @param y      this value will be added to the current one
     */
    public void translateY( double y ) {
        this.translateY += y;
    }

    /**
     * Translates X and Y on the specified value
     *
     * @param x      this value will be added to the current one
     * @param y      this value will be added to the current one
     */
    public void translate( double x, double y ) {
        this.translateX += x;
        this.translateY += y;
    }

    /**
     * Scales current X scale by specified value
     *
     * @param k      multiply the current X scale by this one
     */
    public void scaleX( double k ) {
        this.scaleX *= k;
    }

    /**
     * Scales current Y scale by specified value
     *
     * @param k      multiply the current Y scale by this one
     */
    public void scaleY( double k ) {
        this.scaleY *= k;
    }

    /**
     * Scales current X and Y scales by specified value
     *
     * @param k      multiply the current X and Y scales by this one
     */
    public void scale( double k ) {
        this.scaleX *= k;
        this.scaleY *= k;
    }

    /**
     * Scales current X and Y scales by specified values
     *
     * @param xk      multiply the current X scale by this one
     * @param yk      multiply the current Y scale by this one
     */
    public void scale( double xk, double yk ) {
        this.scaleX *= xk;
        this.scaleY *= yk;
    }

    /**
     * Parses MATRIX tag
     *
     * @param p      parser
     * @return parsed matrix
     */
    public static Matrix parse( Parser p ) {
        p.initBits();

        Matrix m = new Matrix();

        // Scale terms
        if( p.getBool() ) {
            int nBits = p.getBits(5);
            m.setScaleX( Util.fixed2double(p.getSBits(nBits)) );
            m.setScaleY( Util.fixed2double(p.getSBits(nBits)) );
        }
        // Rotate/skew terms
        if( p.getBool() ) {
            int nBits = p.getBits(5);
            m.setRotateSkew0( Util.fixed2double(p.getSBits(nBits)) );
            m.setRotateSkew1( Util.fixed2double(p.getSBits(nBits)) );
        }
        // Translate terms
        int nBits = p.getBits(5);
        m.setTranslateX( p.getSBits(nBits) );
        m.setTranslateY( p.getSBits(nBits) );

        return m;
    }

    /**
     * Skips MATRIX tag without creating Matrix object
     *
     * @param p      parser
     */
    public static void skip( Parser p ) {
        p.initBits();
        // Scale terms
        if( p.getBool() ) {
            int nBits = p.getBits(5);
            p.skipBits(nBits+nBits);
        }
        // Rotate/skew terms
        if( p.getBool() ) {
            int nBits = p.getBits(5);
            p.skipBits(nBits+nBits);
        }
        // Translate terms
        int nBits = p.getBits(5);
        p.skipBits(nBits+nBits);
    }

    public void write( FlashOutput fob ) {
        fob.initBits();
        if( scaleX != 1.0 || scaleY != 1.0 ) {
            fob.writeBit(1);
            int i_scaleX = Util.double2fixed(scaleX);
            int i_scaleY = Util.double2fixed(scaleY);
            int nBits = Util.getMinBitsS( Util.getMax(i_scaleX,i_scaleY) );
            fob.writeBits(nBits, 5);
            fob.writeBits(i_scaleX, nBits);
            fob.writeBits(i_scaleY, nBits);
        } else {
            fob.writeBit(0);
        }
        if( rotateSkew0 != 0.0 || rotateSkew1 != 0.0 ) {
            fob.writeBit(1);
            int i_rotateSkew0 = Util.double2fixed(rotateSkew0);
            int i_rotateSkew1 = Util.double2fixed(rotateSkew1);
            int nBits = Util.getMinBitsS( Util.getMax(i_rotateSkew0,i_rotateSkew1) );
            fob.writeBits(nBits, 5);
            fob.writeBits(i_rotateSkew0, nBits);
            fob.writeBits(i_rotateSkew1, nBits);
        } else {
            fob.writeBit(0);
        }
        int nBits = Util.getMinBitsS( Util.getMax(translateX,translateY) );
        fob.writeBits(nBits, 5);
        fob.writeBits(translateX, nBits);
        fob.writeBits(translateY, nBits);
        fob.flushBits();
    }

    /**
     * Transforms specified Rect by this matrix
     * <p>
     * Does not change specified rectangle, creates new one.
     *
     * @param r      rectangle to be tranformed
     * @return new tranformed rectangle
     */
    public Rect transform( Rect r ) {
        Rect nr = new Rect();
        nr.xmin = (int) (r.xmin*scaleX + r.ymin*rotateSkew1) + translateX;
        nr.ymin = (int) (r.xmin*rotateSkew0 + r.ymin*scaleY) + translateY;
        nr.xmax = (int) (r.xmax*scaleX + r.ymax*rotateSkew1) + translateX;
        nr.ymax = (int) (r.xmax*rotateSkew0 + r.ymax*scaleY) + translateY;
        return nr;
    }

    /**
     * Multiply this matrix by specified one
     * <P>
     * Does not change this matrix nor specified one.
     *
     * @param m      specified matrix
     * @return &lt;new matrix&gt; = &lt;this matrix&gt;*&lt;specified one&gt;
     */
    public Matrix multiply( Matrix m ) {
        Matrix newMat = new Matrix();
        newMat.scaleX      = scaleX        * m.scaleX       +  rotateSkew1   * m.rotateSkew0;
        newMat.rotateSkew1 = scaleX        * m.rotateSkew1  +  rotateSkew1   * m.scaleY;
        newMat.rotateSkew0 = rotateSkew0   * m.scaleX       +  scaleY        * m.rotateSkew0;
        newMat.scaleY      = rotateSkew0   * m.rotateSkew1  +  scaleY        * m.scaleY;
        newMat.translateX  = (int) (m.translateX * scaleX      +  m.translateY * rotateSkew1) + translateX;
        newMat.translateY  = (int) (m.translateX * rotateSkew0 +  m.translateY * scaleY     ) + translateY;
        return newMat;
    }

    public void printContent( PrintStream out, String indent ) {
        out.println( indent+"Matrix: |"+scaleX+"  "+rotateSkew0+"|" );
        out.println( indent+"        |"+rotateSkew1+"  "+scaleY+"|" );
        out.println( indent+"        |"+translateX+"  "+translateY+"|" );
    }

    protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
        super.copyInto( item, copier );
        ((Matrix)item).scaleX      = scaleX;
        ((Matrix)item).scaleY      = scaleY;
        ((Matrix)item).rotateSkew0 = rotateSkew0;
        ((Matrix)item).rotateSkew1 = rotateSkew1;
        ((Matrix)item).translateX  = translateX;
        ((Matrix)item).translateY  = translateY;
        return item;
    }
    public FlashItem getCopy( ScriptCopier copier ) {
        return copyInto( new Matrix(), copier );
    }

}
