/*
 * $Id: Rect.java,v 1.4 2001/07/26 22:56:36 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api;

import java.io.*;
import com.iv.flash.util.*;
import com.iv.flash.parser.*;

/**
 * Rectangle
 * <P>
 * It's defined by it's top-left and bottom-right corners
 *
 * @author Dmitry Skavish
 */
public final class Rect extends FlashItem {

    int xmin;
    int ymin;
    int xmax;
    int ymax;

    public Rect() {}

    /**
     * Creates rectangle specified by it's top-left and bottom-right corners
     *
     * @param xmin   left most X
     * @param ymin   top most Y
     * @param xmax   right most X
     * @param ymax   bottom most Y
     */
    public Rect( int xmin, int ymin, int xmax, int ymax ) {
        this.xmin = xmin;
        this.ymin = ymin;
        this.xmax = xmax;
        this.ymax = ymax;
    }

    /**
     * Creates rectangle
     *
     * @param invalid if true then creates invalid rectangle to be used when union objects' bounds
     */
    public Rect( boolean invalid ) {
        if( invalid ) {
            setXMax(-10000);
            setYMax(-10000);
            setXMin(10000);
            setYMin(10000);
        }
    }

    public int getXMin() {
        return xmin;
    }
    public int getYMin() {
        return ymin;
    }
    public int getXMax() {
        return xmax;
    }
    public int getYMax() {
        return ymax;
    }

    public void setXMin( int v ) {
        xmin = v;
    }
    public void setYMin( int v ) {
        ymin = v;
    }
    public void setXMax( int v ) {
        xmax = v;
    }
    public void setYMax( int v ) {
        ymax = v;
    }

    /**
     * Returns width of this rectangle
     *
     * @return width of the rectangle
     */
    public int getWidth() {
        return xmax-xmin;
    }

    /**
     * Returns height of this rectangle
     *
     * @return height of the rectangle
     */
    public int getHeight() {
        return ymax-ymin;
    }

    public void write( FlashOutput fob ) {
        fob.initBits();
        int nBits = Util.getMinBitsS( Util.getMax(xmin,xmax,ymin,ymax) );
        fob.writeBits( nBits, 5 );
        fob.writeBits( xmin, nBits );
        fob.writeBits( xmax, nBits );
        fob.writeBits( ymin, nBits );
        fob.writeBits( ymax, nBits );
        fob.flushBits();
    }

    public void printContent( PrintStream out, String indent ) {
        out.println( indent+"Rect: ("+xmin+","+ymin+")-("+xmax+","+ymax+")" );
    }

    public static Rect parse( Parser p ) {
        Rect r = new Rect();
        p.initBits();
        int nBits = p.getBits(5);
        r.xmin = p.getSBits(nBits);
        r.xmax = p.getSBits(nBits);
        r.ymin = p.getSBits(nBits);
        r.ymax = p.getSBits(nBits);
        return r;
    }

    public static void skip( Parser p ) {
        p.initBits();
        int nBits = p.getBits(5);
        p.skip( ((5+(nBits*4))+7)/8 - 1  );
    }

    protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
        super.copyInto( item, copier );
        ((Rect)item).xmin = xmin;
        ((Rect)item).ymin = ymin;
        ((Rect)item).xmax = xmax;
        ((Rect)item).ymax = ymax;
        return item;
    }
    public FlashItem getCopy( ScriptCopier copier ) {
        return copyInto( new Rect(), copier );
    }

    /**
     * Returns size of this rectangle in bytes if it 's to be written in flash buffer
     *
     * @return size of this rectangle in bytes
     */
    public int getSize() {
        int nBits = Util.getMinBitsS( Util.getMax(xmin,xmax,ymin,ymax) );
        int s = 5+nBits*4;
        return (s+7)/8;
    }

    /**
     * Adds specified rectangle to this one
     * <P>
     * Does union of two rectangles
     *
     * @param r      specified rectangle
     */
    public void add( Rect r ) {
        if( r == null ) return;
        if( xmin > r.xmin ) xmin = r.xmin;
        if( ymin > r.ymin ) ymin = r.ymin;
        if( xmax < r.xmax ) xmax = r.xmax;
        if( ymax < r.ymax ) ymax = r.ymax;
    }

    /**
     * Creates the same rectangle but defined by opposite corners:
     * bottom-left and top-right (swap Ys)
     *
     * @return opposite rectangle
     */
    public Rect getTrans() {
        Rect r = new Rect();

        r.xmin = xmin;
        r.ymin = ymax;

        r.xmax = xmax;
        r.ymax = ymin;

        return r;
    }

    /**
     * Adds bottom-left, top-right rectangle to this one
     *
     * @param r specified opposite rectangle
     */
    public void addTrans( Rect r ) {
        if( r == null ) return;
        if( xmin > r.xmin ) xmin = r.xmin;
        if( xmin > r.xmax ) xmin = r.xmax;

        if( ymin > r.ymin ) ymin = r.ymin;
        if( ymin > r.ymax ) ymin = r.ymax;

        if( xmax < r.xmin ) xmax = r.xmin;
        if( xmax < r.xmax ) xmax = r.xmax;

        if( ymax < r.ymin ) ymax = r.ymin;
        if( ymax < r.ymax ) ymax = r.ymax;
    }

}

