/*
 * $Id: Script.java,v 1.17 2001/08/22 22:05:57 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api;

import java.io.*;
import com.iv.flash.util.*;
import com.iv.flash.log.*;
import com.iv.flash.parser.*;
import com.iv.flash.commands.*;
import com.iv.flash.api.text.*;
import com.iv.flash.api.image.*;
import com.iv.flash.api.sound.*;
import com.iv.flash.api.shape.*;
import com.iv.flash.api.action.*;
import com.iv.flash.api.button.*;

/**
 * Flash movie clip
 * <p>
 * Contains a timeline, a background color and generator commands of script level.
 * <P>
 * Excerpt from flash file format specs:
 * <P>
 * A script corresponds to a 'movie clip' in the Flash user-interface.
 * It is a movie contained within a movie, and supports many of the features of a regular Flash movie, including:<br>
 * <ul>
 * <li>Most of the control tags that can be used in the main movie.
 * <li>A timeline that can stop, start and play independently of the main movie.
 * <li>A streaming sound track that is automatically mixed with the main sound track.
 * </ul>
 * <p>
 * A script object is defined with a DefineSprite tag.  It consists of a character ID, a frame count, and a
 * series of control tags.  Definition tags (such as DefineShape) are not allowed in the DefineSprite tag.
 * All the characters referred to by control tags in the sprite must be defined outside the sprite,
 * and before the DefineSprite tag.
 * <p>
 * Once defined, a script is displayed with a PlaceObject2 tag in the main movie.  The transform (specified in
 * PlaceObject) is concatenated with the transforms of objects placed inside the script.  These objects behave
 * like 'children' of the script, so when the script is moved, the objects inside the script move also.
 * Similarly, when the script is scaled or rotated, the child objects are also scaled or rotated.
 * A script object stops playing automatically when it is removed from the display list.
 *
 * @author Dmitry Skavish
 */
public final class Script extends FlashDef {

    private Timeline timeline;              // a timeline of the script
    private boolean isProcessed = false;    // true if this script is already processed by jgenerator
    private IVVector gl_commands;           // global generator commands, like SetEnvironment etc
    private IVVector zero_frame;            // zero frame, it usually contains some sound heads etc
    private SetBackgroundColor bkgColor;    // background color of the script

    /**
     * Creates empty script
     */
    public Script() {}

    /**
     * Creates empty script with specified initial capacity
     *
     * @param frameNumber initial capacity of this script in frames
     * @see #Script(int, int)
     */
    public Script( int frameNumber ) {
        this( 0, frameNumber );
    }

    /**
     * Creates empty script with specified initial capacity and ID
     * <P>
     * The capacity specifies the capacity of this script's
     * timeline (in frames). It does not specify how many frames
     * this script will eventually have.
     *
     * @param objID       ID of this script, if it's -1, then this is a main script
     * @param frameNumber initial capacity of this script in frames
     */
    public Script( int objID, int frameNumber ) {
        timeline = new Timeline( frameNumber+1 );
        setID( objID );
    }

    /**
     * Returns true if this script is a main script
     * <P>
     * Main script is a script which represents main flash timeline
     *
     * @return true if this is a main script
     */
    public boolean isMain() {
        return getID() == -1;
    }

    /**
     * Sets this script to be main
     */
    public void setMain() {
        setID( -1 );
    }

    /**
     * Sets this script to be NOT main script
     */
    public void resetMain() {
        setID( 0 );
    }

    /**
     * Returns this script's timeline
     *
     * @return script's timeline
     */
    public Timeline getTimeline() {
        return timeline;
    }

    /**
     * Writes content of this script to flash buffer
     * <P>
     * This method has to be used only if this script is main script.
     *
     * @param fob    flash buffer to write
     * @see #isMain
     * @see #write
     */
    public void generate( FlashOutput fob ) {
        int frameCount = timeline.getFrameCount();
        fob.writeWord( frameCount );

        // collect fonts and make sure there is only one instance of each font
        FontsCollector fc = new FontsCollector();
        collectFonts( fc );

        // write background color first
        if( bkgColor != null ) bkgColor.write( fob );

        // generate timeline in right order (definitions first)
        timeline.generate( fob, new DepsCollector( fc ) );

        Tag.END_TAG.write( fob );
    }

    /**
     * Writes content of this script to flash buffer
     * <P>
     * This method has to be used only if this script is NOT a main script.
     *
     * @param fob    flash buffer to write
     * @see #isMain
     * @see #generate
     */
    public void write( FlashOutput fob ) {
        int start = fob.getPos();     // save for length calculating
        fob.skip( 6 );              // 6 - long tag
        fob.writeDefID( this );
        int frameCount = timeline.getFrameCount();
        fob.writeWord( frameCount );

        // write background color first
        if( bkgColor != null ) bkgColor.write( fob );

        // write "zero frame"
        if( zero_frame != null ) zero_frame.write( fob );

        // write timeline
        timeline.write( fob );

        // write end tag
        Tag.END_TAG.write( fob );

        int size = fob.getPos()-start-6;  // 6 - long tag
        fob.writeLongTagAt( Tag.DEFINESPRITE, size, start );
    }

    /**
     * Parses script
     *
     * @param p      parser
     * @param isMain true if script to be parsed is main
     * @return new parsed script
     * @exception IVException
     */
    public static Script parse( Parser p, boolean isMain ) throws IVException {
        int objID = -1;
        if( !isMain ) {
            objID = p.getUWord();
        }
        // number of frames
        int frameNum = p.getUWord();
        // create script
        Script sc = new Script( objID, frameNum );

        Timeline tl = sc.getTimeline();
        IVVector tFrame = new IVVector();
        sc.zero_frame = tFrame;
        String frameName = null;

        for(;;) {
            int code = p.getTag();

            // check for tag
            switch( code ) {
                /* Control tags         */
                case Tag.SHOWFRAME: {
                    Frame f = new Frame( tFrame );
                    f.setName( frameName );
                    frameName = null;
                    tl.addFrame( f );
                    tFrame.reset();
                    break;
                }
                case Tag.END:
                    return sc;
                case Tag.PLACEOBJECT:
                    tFrame.addElement( Instance.parse( p ) );
                    break;
                case Tag.PLACEOBJECT2:
                    tFrame.addElement( Instance.parse2( p ) );
                    break;
                case Tag.REMOVEOBJECT:
                    tFrame.addElement( RemoveObject.parse( p ) );
                    break;
                case Tag.REMOVEOBJECT2:
                    tFrame.addElement( RemoveObject.parse2( p ) );
                    break;
                case Tag.SETBKGCOLOR:
                    sc.setBackgroundColor( SetBackgroundColor.parse(p) );
                    break;
                case Tag.STARTSOUND:
                    tFrame.addElement( StartSound.parse(p) );
                    break;
                case Tag.DOACTION:
                    tFrame.addElement( DoAction.parse(p) );
                    break;
                case Tag.TEMPLATECOMMAND: {
                    GenericCommand cmd = GenericCommand.parse(p);
                    if( cmd == null ) break;
                    if( (cmd.getType()&0x08) == 0 ) { // check if it is global command
                        FlashItem item = (FlashItem) tFrame.elementAt( tFrame.size()-1 );
                        if( !(item instanceof Instance) ) {
                            Log.log( "genCmd1", cmd.getCommandName() );
                        }
                        cmd.setInstance( (Instance) item );
                        ((Instance)item).setCommand(cmd);
                    } else {
                        sc.addGlobalCommand( cmd );
                    }
                    break;
                }
                /* Info tags            */
                case Tag.FLASHGENERATOR:
                    p.getFile().setTemplate(true);
                    break;
                case Tag.PROTECT:
                    break;
                case Tag.SERIALNUMBER:
                    break;
                case Tag.FRAMELABEL:
                    frameName = p.getString();
                    break;
                case Tag.GENERATORTEXT: {
                    int id = p.getUWord();
                    Text text = (Text) p.getDef(id);
                    text.parseGenText(p);
                    break;
                }
                case Tag.NAMECHARACTER: {
                    int id = p.getUWord();
                    String name = p.getString();
                    FlashDef def = p.getDef(id);
                    def.setName(name);
                    p.addDefToLibrary( name, def );
                    break;
                }
                case Tag.FREECHARACTER:
                    tFrame.addElement( FreeCharacter.parse( p ) );
                    break;
                /* Streaming sound tags */
                case Tag.SOUNDSTREAMHEAD:
                case Tag.SOUNDSTREAMHEAD2:
                    tFrame.addElement( SoundStreamHead.parse( p ) );
                    break;
                case Tag.SOUNDSTREAMBLOCK:
                    tFrame.addElement( SoundStreamBlock.parse( p ) );
                    break;
                /* Definitions tags     */
                case Tag.DEFINESPRITE:
                    p.addDef( Script.parse( p, false ) );
                    break;
                case Tag.DEFINEMOVIE:
                    p.addDef( QTMovie.parse(p) );
                    break;
                case Tag.DEFINESHAPE:
                case Tag.DEFINESHAPE2:
                case Tag.DEFINESHAPE3:
                    p.addDef( LazyShape.parse( p ) );
                    break;
                case Tag.DEFINEMORPHSHAPE:
                    p.addDef( LazyMorphShape.parse( p ) );
                    break;
                case Tag.DEFINESOUND:
                    p.addDef( LazySound.parse( p ) );
                    break;
                case Tag.DEFINEFONT:
                    p.addDef( FontDef.parse( p ) );
                    break;
                case Tag.DEFINEFONTINFO:
                    FontDef.parseInfo( p );
                    break;
                case Tag.DEFINEFONT2:
                    p.addDef( FontDef.parse2( p ) );
                    break;
                case Tag.EXTERNALFONT: {
                    FlashDef def = FontDef.parse3( p );
                    if( def != null ) p.addDef( def );
                    break;
                }
                case Tag.DEFINETEXT:
                    p.addDef( Text.parse( p, false ) );
                    break;
                case Tag.DEFINETEXT2:
                    p.addDef( Text.parse( p, true ) );
                    break;
                case Tag.DEFINEEDITTEXT:
                    p.addDef( TextField.parse( p ) );
                    break;
                case Tag.DEFINEBUTTON:
                    Button b = Button.parse(p);
                    //b.printContent( System.out, "\t" );
                    p.addDef( b );
                    //-- p.addDef( Button.parse( p ) );
                    break;
                case Tag.DEFINEBUTTON2:
                    Button2 b2 = Button2.parse2(p);
                    //b2.printContent( System.out, "\t" );
                    p.addDef( b2 );
                    //--p.addDef( Button2.parse2( p ) );
                    break;
                case Tag.DEFINEBUTTONCXFORM:
                    ButtonCXForm.parse(p);  // this guy will add itself to a button, we need to do nothing
                    break;
                case Tag.DEFINEBUTTONSOUND:
                    ButtonSound.parse(p);  // this guy will add itself to a button, we need to do nothing
                    break;
                case Tag.DEFINEBITS:
                case Tag.DEFINEBITSJPEG2:
                case Tag.DEFINEBITSJPEG3:
                    p.addDef( JPEGBitmap.parse(p) );
                    break;
                case Tag.DEFINEBITSLOSSLESS:
                case Tag.DEFINEBITSLOSSLESS2:
                    p.addDef( LLBitmap.parse(p) );
                    break;
                case Tag.JPEGTABLES:
                    JPEGBitmap.parseJPegTables(p);
                    break;
                    /* Flash 5 tags         */
                case Tag.ENABLEDEBUGGER:
                    System.err.println("ENABLEDEBUGGER!!!");
                    break;
                case 0x64:
                    System.err.println("ENABLEDEBUGGER2!!!");
                    break;
                case Tag.EXPORTASSETS:
                    tFrame.addElement( ExportAssets.parse(p) );
                    break;
                case Tag.IMPORTASSETS:
                    System.err.println("IMPORTASSTES!!!");
                    tFrame.addElement( ImportAssets.parse(p) );
                    break;
                    /* Unknown tags parsed here   */
                default:
                    Log.log( "unknownTag", Util.b2h(code) );
                    tFrame.addElement( p.newUnknownTag() );
                    break;
            }
            p.skipLastTag();
        }
    }

    /**
     * Returns background color of this script
     *
     * @return background color of this script
     */
    public SetBackgroundColor getBackgroundColor() {
        return bkgColor;
    }

    /**
     * Sets new background color
     *
     * @param bkgColor new color
     */
    public void setBackgroundColor( SetBackgroundColor bkgColor ) {
        this.bkgColor = bkgColor;
    }

    protected void addGlobalCommand( GenericCommand cmd ) {
        if( gl_commands == null ) gl_commands = new IVVector();
        gl_commands.addElement( cmd );
    }

    /**
     * Finds nested script's instance by name
     * <P>
     * Searches instance of a script in this script
     * and in all nested scripts by instance's name.
     *
     * @param name   name of the instance to be searched
     * @return found Script or null
     */
    public Instance findInstance( String name ) {
        for( int i=0; i<timeline.getFrameCount(); i++ ) {
            Frame frame = timeline.getFrameAt(i);
            for( int j=0; j<frame.size(); j++ ) {
                FlashObject fo = frame.getFlashObjectAt(j);
                if( !(fo instanceof Instance) ) continue;
                Instance inst = (Instance) fo;
                if( inst.name != null && inst.name.equals( name ) ) return inst;
                if( inst.isScript() ) {
                    inst = inst.getScript().findInstance(name);
                    if( inst != null ) return inst;
                }
            }
        }
        return null;
    }

    /**
     * Processes this script in the specified context and flash file
     *
     * @param file    file to be used when processing this script
     * @param context context to be used when processing this script
     * @exception IVException
     */
    public void process( FlashFile file, Context context ) throws IVException {

        FakeContext fakeContext = null;
        Context localContext = null;

        // execute global commands (if any)
        if( gl_commands != null ) {
            fakeContext = new FakeContext(context);
            for( int i=0; i<gl_commands.size(); i++ ) {
                GenericCommand cmd = (GenericCommand) gl_commands.elementAt(i);
                try {
                    cmd.doCommand(file, localContext!=null?localContext:fakeContext, this, 0);
                    localContext = fakeContext.getContext();
                } catch( Exception e ) {
                    Log.log( new IVException(e, "errDoCmd", new Object[] {
                                file.getFullName(), this.getName(),
                                "0", cmd.getCommandName() } ));
                }
            }
        }

        if( localContext == null ) {
            localContext = new StandardContext();
        }
        localContext.setParent(context);

        // perform generator commands
        timeline.doCommand( file, localContext, this );

        // process inner scripts
        timeline.process( file, localContext );

        // apply context to the rest
        apply( localContext );
    }

    /**
     * Returns true if this script was already processed
     *
     * @return true if this script was already processed
     */
    public boolean isProcessed() {
        return isProcessed;
    }

    /**
     * Marks this script as processed
     */
    public void setProcessed() {
        isProcessed = true;
    }

    /**
     * Creates a copy of this script
     *
     * @return copy of this script
     */
    public Script copyScript() {
        return (Script) getCopy( new ScriptCopier() );
    }

    /**
     * Appends specified script to the end of this script
     * <p>
     * Removes all hanging instances of this script, i.e. for every instance
     * which is still on the timeline by the of this script, put RemoveObject
     * tag in the last frame of the timeline. Then adds all frames of the specified
     * script to the end of this one.
     *
     * @param sc     script to be appended
     */
    public void appendScript( Script sc ) {
        Timeline scTm = sc.getTimeline();

        Frame lastFrame = newFrame();
        removeAllInstances(lastFrame);
        lastFrame.append( scTm.getFrameAt(0) );

        for( int i=1; i<scTm.getFrameCount(); i++ ) {
            Frame frame = scTm.getFrameAt(i);
            timeline.addFrame(frame);
        }
    }

    /**
     * Removes all hanging instances of the timeline in the specified frame
     * <p>
     * Traverses this script and for every hanging instance (instance which is
     * still on the timeline by the last frame) puts RemoveObject tag in the specified frame.
     *
     * @param lastFrame frame to put RemoveObject tags in
     */
    public void removeAllInstances( Frame lastFrame ) {
        IVVector layers = new IVVector();
        for( int i=0; i<timeline.getFrameCount(); i++ ) {
            Frame frame = timeline.getFrameAt(i);
            for( int k=0; k<frame.size(); k++ ) {
                FlashObject fo = frame.getFlashObjectAt(k);
                if( fo instanceof Instance ) {
                    Instance inst = (Instance) fo;
                    int layer = inst.depth;
                    layers.setElementAt(inst, layer);
                } else if( fo instanceof RemoveObject ) {
                    RemoveObject ro = (RemoveObject) fo;
                    int layer = ro.depth;
                    layers.setElementAt(null, layer);
                }
            }
        }
        for( int i=0; i<layers.size(); i++ ) {
            Instance inst = (Instance) layers.elementAt(i);
            if( inst == null ) continue;
            RemoveObject ro = new RemoveObject();
            ro.depth = i;
            lastFrame.addFlashObject( ro );
        }
    }

    /**
     * Calculates bounds of this script
     *
     * @return bounds of this script
     */
    public Rect getBounds() {
        Rect rect = new Rect(true);
        IVVector layers = new IVVector();
        for( int i=0; i<timeline.getFrameCount(); i++ ) {
            Frame frame = timeline.getFrameAt(i);
            for( int k=0; k<frame.size(); k++ ) {
                FlashObject fo = frame.getFlashObjectAt(k);
                if( fo instanceof Instance ) {
                    Instance inst = (Instance) fo;
                    int layer = inst.depth;
                    Rect bounds = null;
                    FlashDef def = inst.def;
                    if( def != null ) {
                        bounds = def.getBounds();
                        layers.setElementAt(bounds, layer);
                    } else if( inst.matrix != null ) {
                        bounds = (Rect) layers.elementAt(layer);
                    }
                    if( inst.matrix != null && bounds != null ) {
                        Rect trans = bounds.getTrans();
                        bounds = inst.matrix.transform(bounds);
                        trans = inst.matrix.transform(trans);
                        bounds.addTrans(trans);
                    }
                    rect.add(bounds);
                } else if( fo instanceof RemoveObject ) {
                    RemoveObject ro = (RemoveObject) fo;
                    int layer = ro.depth;
                    layers.setElementAt(null, layer);
                } else {
                    rect.add( fo.getBounds() );
                }
            }
        }
        return rect;
    }

    /**
     * Reserves specified number of layers beginning from the specified one
     * <p>
     * Traverses the timeline and moves all the instances which are on layers
     * greater than the specified one by the specified number.
     * <P>
     * For example if we have instances A, B, C, D on layers 2,3,4,5,
     * then if we call reserveLayers(3,2) we will have:
     * A on 2, B on 5, C on 6, D on 7, so there will be two available layers depths: 3 and 4.
     *
     * @param from   reserve layers beginning from this one
     * @param num    number of layers to be reserved
     * @return 'from' layer
     */
    public int reserveLayers( int from, int num ) {
        for( int i=0; i<timeline.getFrameCount(); i++ ) {
            Frame frame = timeline.getFrameAt(i);
            for( int k=0; k<frame.size(); k++ ) {
                FlashObject fo = frame.getFlashObjectAt(k);
                if( fo instanceof Instance ) {
                    Instance inst = (Instance) fo;
                    if( inst.depth >= from ) inst.depth += num;
                    if( inst.clip >= from ) inst.clip += num;
                } else if( fo instanceof RemoveObject ) {
                    RemoveObject ro = (RemoveObject) fo;
                    if( ro.depth >= from ) ro.depth += num;
                }
            }
        }
        return from;
    }

    /**
     * Returns maximum layer's depth of this script
     *
     * @return maximum depth
     */
    public int getMaxDepth() {
        int max = 0;
        for( int i=0; i<timeline.getFrameCount(); i++ ) {
            Frame frame = timeline.getFrameAt(i);
            for( int k=0; k<frame.size(); k++ ) {
                FlashObject fo = frame.getFlashObjectAt(k);
                if( fo instanceof Instance ) {
                    Instance inst = (Instance) fo;
                    if( inst.depth > max ) max = inst.depth;
                }
            }
        }
        return max;
    }

    /**
     * Returns number of frames in this script
     *
     * @return number of frames
     */
    public int getFrameCount() {
        return timeline.getFrameCount();
    }

    /**
     * Returns frame at specified index
     *
     * @param frame  frame number
     * @return frame at specified index
     */
    public Frame getFrameAt( int frame ) {
        return timeline.getFrameAt(frame);
    }

    /**
     * Creates new frame and adds it to the end of the timeline
     *
     * @return new added frame
     */
    public Frame newFrame() {
        return timeline.newFrame();
    }

    /**
     * Returns last frame of this script
     *
     * @return last frame of this script
     */
    public Frame getLastFrame() {
        return timeline.getFrameAt( getFrameCount()-1 );
    }

    public void apply( Context context ) {
        if( isConstant() || isProcessed() ) return;
        timeline.apply(context);
    }

    public boolean isConstant() {
        return timeline.isConstant();
    }

    protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
        super.copyInto( item, copier );
        ((Script)item).zero_frame = zero_frame!=null? zero_frame.getCopy(copier): null;
        ((Script)item).timeline = (Timeline) timeline.getCopy(copier);
        ((Script)item).isProcessed = isProcessed;
        if( gl_commands != null ) {
            IVVector v = new IVVector( gl_commands.size() );
            for( int i=0; i<gl_commands.size(); i++ ) {
                GenericCommand cmd = (GenericCommand) gl_commands.elementAt(i);
                v.addElement( cmd.getCopy(null) );
            }
            ((Script)item).gl_commands = v;
        }
        return item;
    }

    public FlashItem getCopy( ScriptCopier copier ) {
        return copyInto( new Script(), copier );
    }

    public void collectDeps( DepsCollector dc ) {
        timeline.collectDeps(dc);
    }

    public void collectFonts( FontsCollector fc ) {
        for( int i=0; i<timeline.getFrameCount(); i++ ) {
            Frame frame = timeline.getFrameAt(i);
            for( int k=0; k<frame.size(); k++ ) {
                FlashObject fo = frame.getFlashObjectAt(k);
                fo.collectFonts( fc );
            }
        }
    }

    public int getTag() {
        return Tag.DEFINESPRITE;
    }

    public void printContent( PrintStream out, String indent ) {
        String id = isMain()? "main": Integer.toString(getID());
        out.println( indent+"Script: id="+id+" frames="+timeline.getFrameCount()+" name='"+getName()+"'" );
        if( zero_frame != null ) zero_frame.printContent( out, indent+"    " );
        timeline.printContent( out, indent+"    " );
        out.println( indent+"End Script("+id+") name='"+getName()+"'" );
    }

}
