/*
 * $Id: FillStyle.java,v 1.2 2001/08/16 04:16:42 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api.shape;

import com.iv.flash.util.*;
import java.io.PrintStream;
import com.iv.flash.api.*;
import com.iv.flash.api.image.*;

/**
 * FillStyle of Shape.
 * <P>
 * SWF supports three basic types of fills for a shape.
 * <UL>
 * <LI>Solid fill - A simple RGBA color that fills a portion of a shape.
 *     An alpha value of 255 means a completely opaque fill.
 *     An alpha value of zero means a completely transparent fill.
 *     Any alpha between 0 and 255 will be partially transparent.
 * <LI>Gradient Fill - A gradient fill can be either a linear or a radial gradient.
 *     See {@link com.iv.flash.api.Gradient}s for an in depth description of how gradients are defined.
 * <LI>Bitmap fill - Bitmap fills refer to a bitmap characterId.
 *     There are two styles: clipped and tiled. A clipped bitmap fill repeats the color on
 *     the edge of a bitmap if the fill extends beyond the edge of the bitmap.
 *     A tiled fill repeats the bitmap if the fill extends beyond the edge of the bitmap.
 * </UL>
 *
 * @author Dmitry Skavish
 */
public final class FillStyle extends FlashItem {

    public static final int SOLID           = 0x00;
    public static final int LINEAR_GRADIENT = 0x10;
    public static final int RADIAL_GRADIENT = 0x12;
    public static final int TILED_BITMAP    = 0x40;
    public static final int CLIPPED_BITMAP  = 0x41;

    private int type;               // type of this fillstyle
    private AlphaColor color;       // color of this fillstyle
    private Matrix matrix;          // matrix of this fillstyle
    private Gradient gradient;      // gradient of this fillstyle
    private Bitmap bitmap;          // bitmap of this fillstyle

    public FillStyle() {}

    public int getType() {
        return type;
    }
    public void setType( int type ) {
        this.type = type;
    }

    public AlphaColor getColor() {
        return color;
    }
    public void setColor( AlphaColor color ) {
        this.color = color;
    }

    public Matrix getMatrix() {
        return matrix;
    }
    public void setMatrix( Matrix matrix ) {
        this.matrix = matrix;
    }

    public Gradient getGraduent() {
        return gradient;
    }
    public void setGradient( Gradient gradient ) {
        this.gradient = gradient;
    }

    public Bitmap getBitmap() {
        return bitmap;
    }
    public void setBitmap( Bitmap bitmap ) {
        this.bitmap = bitmap;
    }

    /**
     * Creates new solid fill.
     *
     * @param color  color of solid fill
     * @return solid fillstyle
     */
    public static FillStyle newSolid( AlphaColor color ) {
        FillStyle fs = new FillStyle();
        fs.setColor( color );
        fs.setType( SOLID );
        return fs;
    }

    /**
     * Creates new linear gradient fill.
     *
     * @param gradient gradient of gradient fill
     * @param matrix   matrix of gradient fill
     * @return linear gradient fill
     */
    public static FillStyle newLinearGradient( Gradient gradient, Matrix matrix ) {
        FillStyle fs = new FillStyle();
        fs.setGradient( gradient );
        fs.setMatrix( matrix );
        fs.setType( LINEAR_GRADIENT );
        return fs;
    }

    /**
     * Creates new radial gradient fill.
     *
     * @param gradient gradient of gradient fill
     * @param matrix   matrix of gradient fill
     * @return radial gradient fill
     */
    public static FillStyle newRadialGradient( Gradient gradient, Matrix matrix ) {
        FillStyle fs = new FillStyle();
        fs.setGradient( gradient );
        fs.setMatrix( matrix );
        fs.setType( RADIAL_GRADIENT );
        return fs;
    }

    /**
     * Creates new tiled bitmap fill.
     *
     * @param bitmap bitmap of bitmap fill
     * @param matrix matrix of bitmap fill
     * @return tiled bitmap fill
     */
    public static FillStyle newTiledBitmap( Bitmap bitmap, Matrix matrix ) {
        FillStyle fs = new FillStyle();
        fs.setBitmap( bitmap );
        fs.setMatrix( matrix );
        fs.setType( TILED_BITMAP );
        return fs;
    }

    /**
     * Creates new clipped bitmap fill.
     *
     * @param bitmap bitmap of bitmap fill
     * @param matrix matrix of bitmap fill
     * @return clipped bitmap fill
     */
    public static FillStyle newClippedBitmap( Bitmap bitmap, Matrix matrix ) {
        FillStyle fs = new FillStyle();
        fs.setBitmap( bitmap );
        fs.setMatrix( matrix );
        fs.setType( CLIPPED_BITMAP );
        return fs;
    }

    /**
     * Creates new tiled bitmap fill with default tranformation matrix.
     * <p>
     * Default matrix translates from pixels space to twixels space.
     *
     * @param bitmap bitmap of bitmap fill
     * @return tiled bitmap fill
     */
    public static FillStyle newTiledBitmap( Bitmap bitmap ) {
        Matrix matrix = new Matrix();
        matrix.setScaleX(20.0);
        matrix.setScaleY(20.0);
        return newTiledBitmap( bitmap, matrix );
    }

    /**
     * Creates new clipped bitmap fill with default tranformation matrix.
     * <p>
     * Default matrix translates from pixels space to twixels space.
     *
     * @param bitmap bitmap of bitmap fill
     * @return clipped bitmap fill
     */
    public static FillStyle newClippedBitmap( Bitmap bitmap ) {
        Matrix matrix = new Matrix();
        matrix.setScaleX(20.0);
        matrix.setScaleY(20.0);
        return newClippedBitmap( bitmap, matrix );
    }

    public void write( FlashOutput fob ) {
        fob.writeByte(type);
        switch( type ) {
            case SOLID:
                color.write(fob);
                break;
            case LINEAR_GRADIENT:
            case RADIAL_GRADIENT:
                matrix.write(fob);
                gradient.write(fob);
                break;
            case TILED_BITMAP:
            case CLIPPED_BITMAP:
                fob.writeDefID( bitmap );
                matrix.write(fob);
                break;
        }
    }

    public void printContent( PrintStream out, String indent ) {
        switch( type ) {
            case SOLID:
                out.println( indent+"FillStyle (SOLID):" );
                color.printContent(out, indent+"    ");
                break;
            case LINEAR_GRADIENT:
                out.println( indent+"FillStyle (LINEAR_GRADIENT):" );
                matrix.printContent(out, indent+"    ");
                gradient.printContent(out, indent+"    ");
                break;
            case RADIAL_GRADIENT:
                out.println( indent+"FillStyle (RADIAL_GRADIENT):" );
                matrix.printContent(out, indent+"    ");
                gradient.printContent(out, indent+"    ");
                break;
            case TILED_BITMAP:
                out.println( indent+"FillStyle (TILED_BITMAP):" );
                matrix.printContent(out, indent+"    ");
                out.println( indent+"    bitmapID="+bitmap.getID() );
                break;
            case CLIPPED_BITMAP:
                out.println( indent+"FillStyle (CLIPPED_BITMAP):" );
                matrix.printContent(out, indent+"    ");
                out.println( indent+"    bitmapID="+bitmap.getID() );
                break;
        }
    }

    protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
        ((FillStyle)item).type = type;
        switch( type ) {
            case SOLID:
                ((FillStyle)item).color = (AlphaColor) color.getCopy(copier);
                break;
            case LINEAR_GRADIENT:
            case RADIAL_GRADIENT:
                ((FillStyle)item).gradient = (Gradient) gradient.getCopy(copier);
                ((FillStyle)item).matrix = (Matrix) matrix.getCopy(copier);
                break;
            case TILED_BITMAP:
            case CLIPPED_BITMAP:
                ((FillStyle)item).bitmap = (Bitmap) copier.copy(bitmap);
                ((FillStyle)item).matrix = (Matrix) matrix.getCopy(copier);
                break;
        }
        return item;
    }

    public FlashItem getCopy( ScriptCopier copier ) {
        return copyInto( new FillStyle(), copier );
    }
}
