/*
 * $Id: Shape.java,v 1.4 2001/08/16 04:16:42 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api.shape;

import com.iv.flash.parser.*;
import com.iv.flash.util.*;
import com.iv.flash.api.*;
import java.io.PrintStream;

/**
 * Shape.
 * <P>
 * Shapes are defined by a list of edges called a path. A path may be closed - where the start
 * and end of the path meet to close the figure, or open - where the path forms an
 * open-ended stroke. A path may contain a mixture of straight edges, curved edges, and
 * "pen up and move" commands.  The latter allows multiple disconnected figures to be
 * described by a single shape structure. (see MoveTo flag)
 * <P>
 * A fill style defines the appearance of an area enclosed by a path. Fill styles supported
 * by SWF include a color, a gradient, or a bitmapped image.<BR>
 * A line style defines the appearance of the outline of a path. The line style may be a stroke
 * of any thickness and color.
 * <P>
 * SWF format allows each edge to have its own line and fill style.
 * This can have unpredictable results when fill styles change in the middle of a path.
 * <P>
 * Flash also supports two fill styles per edge, one for each side of the edge:
 * FillStyle0 and FillStyle1. FillStyle0 should always be used first and then
 * FillStyle1 if the shape is filled on both sides of the edge.
 * <P>
 * A shape is comprised of the following elements:
 * <UL>
 * <LI>CharacterId - A 16-bit value that uniquely identifies this shape as a
 *     'character' in the dictionary.  The CharacterId can be referred to in
 *     control tags such as PlaceObject. Characters can be re-used and combined
 *     with other characters to make more complex shapes.
 * <LI>Bounding box - The rectangle that completely encloses the shape.
 * <LI>Fill style array - A list of all the fill styles used in a shape.
 * <LI>Line style array - A list of all the line styles used in a shape.
 * <LI>Shape-record array - A list of shape-records.  Shape-records can define straight
 *     or curved edges, style changes, or move the drawing position.
 * </UL>
 * Note: Line and fill styles are defined once only, and may be used (and re-used) by
 * any of the edges in the shape.
 * <p>
 * Note that objects of this class will always be generated as DefineShape3 tags which
 * means that all the colors used in the shape (in line and fillstyles) have to be AlphaColors!
 *
 * @author Dmitry Skavish
 */
public final class Shape extends FlashDef {

    private ShapeStyle shapeStyle = new ShapeStyle();   // collection of fill and line styles
    private IVVector shapeRecords = new IVVector();     // collection of shape records
    private Rect bounds;                                // bounding box of this shape

    // temp variables used to track pen movement
    private int lastX = Integer.MAX_VALUE;
    private int lastY = Integer.MAX_VALUE;

    /**
     * Creates new empty shape
     */
    public Shape() {}

    /**
     * Returns vector of shape records.
     *
     * @return vector of shape records
     */
    public IVVector getShapeRecords() {
        return shapeRecords;
    }

    /**
     * Returns array of shape styles
     *
     * @return object representing array of shape styles
     */
    public ShapeStyle getShapeStyle() {
        return shapeStyle;
    }

    /**
     * Adds specified fill style to the array of shape styles.
     * <p>
     * Note: colors used in the specified fillstyle have to be AlphaColors
     *
     * @param fillStyle specified fill style
     * @return index of added fill style in the array
     */
    public int addFillStyle( FillStyle fillStyle ) {
        return shapeStyle.addFillStyle( fillStyle );
    }

    /**
     * Adds specified line style to the array of shape styles
     * <p>
     * Note: colors used in the specified linestyle have to be AlphaColors
     *
     * @param lineStyle specified line style
     * @return index of added line style in the array
     */
    public int addLineStyle( LineStyle lineStyle ) {
        return shapeStyle.addLineStyle( lineStyle );
    }

    /**
     * Draws curve record.
     *
     * @param cx     X control point
     * @param cy     Y control point
     * @param ax     X anchor point
     * @param ay     Y anchor point
     */
    public void curveTo( int cx, int cy, int ax, int ay ) {
        shapeRecords.addElement( new CurvedEdgeRecord(cx-lastX,cy-lastY,ax-cx,ay-cy) );
        lastX = ax;
        lastY = ay;
    }

    /**
     * Draws a straight line from current position to the specified one.
     *
     * @param x      X coordinate draw the line to
     * @param y      Y coordinate draw the line to
     */
    public void lineTo( int x, int y ) {
        int dx = x-lastX;
        int dy = y-lastY;
        if( dx == 0 ) {
            shapeRecords.addElement( StrightEdgeRecord.newVLine(dy) );
        } else if( dy == 0 ) {
            shapeRecords.addElement( StrightEdgeRecord.newHLine(dx) );
        } else {
            shapeRecords.addElement( StrightEdgeRecord.newLine(dx,dy) );
        }
        lastX = x;
        lastY = y;
    }

    /**
     * Draws a straight line specified by its both ends.
     * <P>
     * Moves pen to start of the line (if needed) and draws
     * the line to its end.
     *
     * @param x1     X of the beginning of the line
     * @param y1     Y of the beginning of the line
     * @param x2     X of the end of the line
     * @param y2     Y of the end of the line
     */
    public void line( int x1, int y1, int x2, int y2 ) {
        if( lastX != x1 || lastY != y1 ) moveTo( x1, y1 );
        lineTo( x2, y2 );
    }

    /**
     * Draws a rectangle specified by its top-left and bottom-right corners.
     * <p>
     * Draws up to four lines.
     *
     * @param x1     leftmost X
     * @param y1     topmost Y
     * @param x2     rightmost X
     * @param y2     bottommost Y
     */
    public void rectangle( int x1, int y1, int x2, int y2 ) {
        line( x1, y1, x2, y1 );
        line( x2, y1, x2, y2 );
        line( x2, y2, x1, y2 );
        line( x1, y2, x1, y1 );
    }

    /**
     * Draws a rectangle specified by {@link Rect} structure.
     * <p>
     * Draws up to four lines.
     *
     * @param r      specified rectangle
     */
    public void rectangle( Rect r ) {
        rectangle( r.getXMin(), r.getYMin(), r.getXMax(), r.getYMax() );
    }

    /**
     * Moves pen to the specified position.
     *
     * @param x      new current X
     * @param y      new current Y
     */
    public void moveTo( int x, int y ) {
        StyleChangeRecord sc = getStyleChange();
        sc.addFlags( StyleChangeRecord.MOVETO );
        sc.setDeltaX(x);
        sc.setDeltaY(y);
        lastX = x;
        lastY = y;
    }

    /**
     * Sets specified fillstyle as current fillstyle0.
     * <P>
     * If the specified fillstyle is not in the shapestyle array
     * then it's added, if it's already there it's reused.
     * <p>
     * Note: colors used in the specified fillstyle have to be AlphaColors
     *
     * @param fillStyle specified fillstyle
     * @return index of specified fillstyle in the shapestyle array
     */
    public int setFillStyle0( FillStyle fillStyle ) {
        int fs = shapeStyle.getFillStyleIndex( fillStyle );
        if( fs == -1 ) fs = addFillStyle( fillStyle );
        setFillStyle0( fs );
        return fs;
    }

    /**
     * Sets specified fillstyle as current fillstyle1.
     * <P>
     * If the specified fillstyle is not in the shapestyle array
     * then it's added, if it's already there it's reused.
     * <p>
     * Note: colors used in the specified fillstyle have to be AlphaColors
     *
     * @param fillStyle specified fillstyle
     * @return index of specified fillstyle in the shapestyle array
     */
    public int setFillStyle1( FillStyle fillStyle ) {
        int fs = shapeStyle.getFillStyleIndex( fillStyle );
        if( fs == -1 ) fs = addFillStyle( fillStyle );
        setFillStyle1( fs );
        return fs;
    }

    /**
     * Sets specified linestyle as current linestyle.
     * <P>
     * If the specified linestyle is not in the shapestyle array
     * then it's added, if it's already there it's reused.
     * <p>
     * Note: color used in the specified linestyle has to be AlphaColor
     *
     * @param lineStyle specified linestyle
     * @return index of specified linestyle in the shapestyle array
     */
    public int setLineStyle( LineStyle lineStyle ) {
        int ls = shapeStyle.getLineStyleIndex( lineStyle );
        if( ls == -1 ) ls = addLineStyle( lineStyle );
        setLineStyle( ls );
        return ls;
    }

    /**
     * Sets current fillstyle0 by its index in shapestyle array.
     *
     * @param fillStyle index of fillstyle in shapestyle array to be set as current fillstyle0
     */
    public void setFillStyle0( int fillStyle ) {
        StyleChangeRecord sc = getStyleChange();
        sc.addFlags( StyleChangeRecord.FILLSTYLE0 );
        sc.setFillStyle0( fillStyle );
    }

    /**
     * Sets current fillstyle1 by its index in shapestyle array.
     *
     * @param fillStyle index of fillstyle in shapestyle array to be set as current fillstyle1
     */
    public void setFillStyle1( int fillStyle ) {
        StyleChangeRecord sc = getStyleChange();
        sc.addFlags( StyleChangeRecord.FILLSTYLE1 );
        sc.setFillStyle1( fillStyle );
    }

    /**
     * Sets current linestyle by its index in shapestyle array.
     *
     * @param lineStyle index of linestyle in shapestyle array to be set as current linestyle
     */
    public void setLineStyle( int lineStyle ) {
        StyleChangeRecord sc = getStyleChange();
        sc.addFlags( StyleChangeRecord.LINESTYLE );
        sc.setLineStyle( lineStyle );
    }

    private StyleChangeRecord getStyleChange() {
        if( shapeRecords.size() > 0 ) {
            FlashItem item = (FlashItem) shapeRecords.elementAt( shapeRecords.size()-1 );
            if( item instanceof StyleChangeRecord ) return (StyleChangeRecord) item;
        }
        StyleChangeRecord sc = new StyleChangeRecord();
        shapeRecords.addElement(sc);
        return sc;
    }

    public int getTag() {
        return Tag.DEFINESHAPE3;
    }

    public void collectDeps( DepsCollector dc ) {
        shapeStyle.collectDeps(dc);
    }

    public void write( FlashOutput main ) {
        FlashOutput fob = new FlashOutput(main,100);
        bounds.write(fob);
        shapeStyle.write(fob);
        int nFillBits = shapeStyle.calcNFillBits();
        int nLineBits = shapeStyle.calcNLineBits();
        fob.writeByte( (nFillBits<<4) | nLineBits );
        fob.initBits();
        for( int i=0; i<shapeRecords.size(); i++ ) {
            FlashItem item = (FlashItem) shapeRecords.elementAt(i);
            if( item instanceof StyleChangeRecord ) {
                StyleChangeRecord scr = (StyleChangeRecord) item;
                scr.write(fob, nFillBits, nLineBits);
            } else {
                item.write(fob);
            }
        }
        fob.writeBits(0, 6);
        fob.flushBits();

        main.writeTag( getTag(), 2+fob.getSize() );
        main.writeDefID( this );
        main.writeFOB( fob );
    }

    public void printContent( PrintStream out, String indent ) {
        out.println( indent+"Shape("+Tag.tagNames[getTag()]+"): id="+getID()+", name='"+getName()+"'" );
        bounds.printContent( out, indent+"   " );
        shapeStyle.printContent(out, indent+"    " );
        for( int i=0; i<shapeRecords.size(); i++ ) {
            FlashItem item = (FlashItem) shapeRecords.elementAt(i);
            item.printContent(out, indent+"   ");
        }
    }

    public boolean isConstant() {
        return true;
    }

    public Rect getBounds() {
        return bounds;
    }

    /**
     * Sets bounding box for this shape.
     *
     * @param bounds new bounding box
     */
    public void setBounds( Rect bounds ) {
        this.bounds = bounds;
    }

    protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
        super.copyInto( item, copier );
        ((Shape)item).bounds = (Rect) bounds.getCopy(copier);
        ((Shape)item).shapeStyle = (ShapeStyle) shapeStyle.getCopy(copier);
        ((Shape)item).shapeRecords = shapeRecords.getCopy(copier);
        return item;
    }

    public FlashItem getCopy( ScriptCopier copier ) {
        return copyInto( new Shape(), copier );
    }

    private static Shape emptyShape;

    public synchronized static Shape getEmptyShape() {
        if( emptyShape != null ) return emptyShape;
        emptyShape = new Shape();
        emptyShape.setLineStyle( new LineStyle(1, new AlphaColor(0,0,0,0) ) );
        emptyShape.moveTo(0,0);
        emptyShape.lineTo(0,0);
        emptyShape.setBounds( new Rect(0,0,0,0) );
        return emptyShape;
    }
}
