/*
 * $Id: LazyText.java,v 1.4 2001/08/10 04:55:25 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api.text;

import com.iv.flash.parser.*;
import com.iv.flash.util.*;
import com.iv.flash.api.*;
import java.io.*;
import java.util.*;

/**
 * This class represents DefineText tag.
 * <P>
 * Basically it is alredy laid out text.
 *
 * @author Dmitry Skavish
 * @see Text
 */
public final class LazyText extends FlashDef implements TextBlock {

    public static final int HAS_FONT    = 0x08;
    public static final int HAS_COLOR   = 0x04;
    public static final int HAS_YOFFSET = 0x02;
    public static final int HAS_XOFFSET = 0x01;

    public boolean withAlpha;                   // true if color with alpha
    public Rect bounds;                         // bounds of the text
    public Matrix matrix;                       // matrix of the text
    public IVVector records = new IVVector();   // vector of TextRecord's and TextStyleChangeRecord's

    public LazyText() {}

    public LazyText( boolean withAlpha )    {
        this.withAlpha = withAlpha;
    }

    public Rect getBounds() {
        return bounds;
    }

    public void setBounds( Rect bounds ) {
        this.bounds = bounds;
    }

    /**
     * Returns vector of all text records of this text.
     *
     * @return vector of {@link TextRecord} or/and {@link TextStyleChangeRecord}
     */
    public IVVector getTextRecords() {
        return records;
    }

    /**
     * Sets new vector of text records for this text.
     *
     * @param records  new vector of {@link TextItem} or/and {@link TextStyleChangeRecord}
     */
    public void setTextRecords( IVVector records ) {
        this.records = records;
    }

    /**
     * Adds new text record to this text
     *
     * @param record   new text record to be added
     */
    public void addTextRecord( TextRecord record ) {
        records.addElement(record);
    }

    /**
     * Adds new text style change record to this text
     *
     * @param record   new text style change record to be added
     */
    public void addTextStyleChangeRecord( TextStyleChangeRecord record ) {
        records.addElement(record);
    }

    public void setMatrix( Matrix matrix ) {
        this.matrix = matrix;
    }

    public Matrix getMatrix() {
        return matrix;
    }

    public int getTag() {
        if( withAlpha ) return Tag.DEFINETEXT2;
        return Tag.DEFINETEXT;
    }

    public static LazyText parse( Parser p, boolean withAlpha ) {
        LazyText text = new LazyText( withAlpha );
        text._parse( p );
        return text;
    }

    public void _parse( Parser p ) {
        // get id
        setID( p.getUWord() );
        // get bounds and matrix
        bounds = Rect.parse(p);
        matrix = Matrix.parse(p);

        // get nBits
        int nGlyphBits = p.getUByte();
        int nAdvanceBits = p.getUByte();

        // get textrecords
        Font curFont = null;
        for(;;) {
            int flags = p.getUByte();
            if( flags == 0 ) {
                break;
            }
            if( (flags&0x80) != 0 ) {
                // control record
                TextStyleChangeRecord ts = new TextStyleChangeRecord();
                records.addElement(ts);
                if( (flags&HAS_FONT) != 0 ) {
                    // get font definition
                    curFont = ((FontDef)p.getDef(p.getUWord())).getFont();
                    ts.setFont( curFont );
                }
                if( (flags&HAS_COLOR) != 0 ) {
                    // get color
                    ts.setColor( Color.parse(p, withAlpha) );
                }
                if( (flags&HAS_XOFFSET) != 0 ) {
                    ts.setX( p.getWord() );
                }
                if( (flags&HAS_YOFFSET) != 0 ) {
                    ts.setY( p.getWord() );
                }
                if( (flags&HAS_FONT) != 0 ) {
                    ts.setHeight( p.getUWord() );
                }
            } else {
                TextRecord tr = new TextRecord( flags );    // flags contains number of glyphs
                records.addElement(tr);
                // glyph record
                p.initBits();
                for( int i=0; i<flags; i++ ) {
                    int idx = p.getBits(nGlyphBits);
                    int adv = p.getSBits(nAdvanceBits);

                    char ch = (char) curFont.codeTable[idx];   // codetable has to be already setup                    
                    tr.add(ch, idx, adv);
                }
            }
        }
    }

    public void write( FlashOutput fob ) {
        int start = fob.getPos();     // save for tag
        fob.skip( 6 );    // 6 - long tag
        fob.writeDefID( this );
        bounds.write(fob);
        matrix.write(fob);

        // update indexes from font and calculate maximum index and advance value
        int maxIdx = Integer.MIN_VALUE;
        int maxAdv = Integer.MIN_VALUE;
        Font lastFont = null;
        for( int i=0; i<records.size(); i++ ) {
            Object o = records.elementAt(i);
            if( o instanceof TextStyleChangeRecord ) {
                Font f = ((TextStyleChangeRecord)o).getFont();
                if( f != null ) lastFont = f;
            } else {
                TextRecord tr = (TextRecord) o;
                //tr.updateRecord(lastFont);        !!!!!!!!!!!!!!!!!!!!!!!! UNCOMMENT
                int idx = tr.getMaxIndex();
                int adv = tr.getMaxAdvance();
                if( idx > maxIdx ) maxIdx = idx;
                if( adv > maxAdv ) maxAdv = adv;
            }

        }
        int nGlyphBits = Util.getMinBitsU(maxIdx);
        int nAdvanceBits = Util.getMinBitsS(maxAdv);

        fob.writeByte(nGlyphBits);
        fob.writeByte(nAdvanceBits);

        fob.setUserData( new int[] {nGlyphBits, nAdvanceBits} );
        records.write( fob );

        fob.writeByte(0);

        int size = fob.getPos()-start-6;  // 6 - long tag
        fob.writeLongTagAt( getTag(), size, start );
    }

    public void printContent( PrintStream out, String indent ) {
        out.println( indent+"LazyText: id="+getID() );
        bounds.printContent(out, indent+"    ");
        matrix.printContent(out, indent+"    ");
        records.printContent( out, indent+"    " );
    }

    // ------------------------------------------------------- //
    //                 TextBlock implementation                //
    /**
     * Layouts this text.
     * <p>
     * Does nothing, because the text is already laid out.
     */
    public void layout() {
    }

    /**
     * Returns vector of {@link TextRecord}s from this text
     * of specified font.
     *
     * @param font   font of text records to be returned
     * @return text records of specified font
     */
    public IVVector getTextRecords( Font font ) {
        IVVector trs = new IVVector();
        Font lastFont = null;
        for( int k=0; k<records.size(); k++ ) {
            Object o = records.elementAt(k);
            if( o instanceof TextStyleChangeRecord ) {
                Font f = ((TextStyleChangeRecord)o).getFont();
                if( f != null ) lastFont = f;
            } else {
                if( lastFont == font ) trs.addElement(o);
            }
        }
        return trs;
    }
    // ------------------------------------------------------- //

    public void collectDeps( DepsCollector dc ) {
        for( int i=0; i<records.size(); i++ ) {
            FlashItem t = (FlashItem) records.elementAt(i);
            if( t instanceof TextStyleChangeRecord ) {
                TextStyleChangeRecord ts = (TextStyleChangeRecord) t;
                if( ts.getFont() != null ) dc.addDep(ts.getFont());
            }
        }
    }

    public void collectFonts( FontsCollector fc ) {
        for( int i=0; i<records.size(); i++ ) {
            FlashItem t = (FlashItem) records.elementAt(i);
            if( t instanceof TextStyleChangeRecord ) {
                TextStyleChangeRecord ts = (TextStyleChangeRecord) t;
                if( ts.getFont() != null ) fc.addFont( ts.getFont(), this );
            }
        }
    }

    public boolean isConstant() {
        return true;
    }

    protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
        super.copyInto( item, copier );
        ((LazyText)item).withAlpha = withAlpha;
        ((LazyText)item).bounds = (Rect) bounds.getCopy(copier);
        ((LazyText)item).matrix = (Matrix) matrix.getCopy(copier);
        ((LazyText)item).records = records.getCopy(copier);
        return item;
    }

    public FlashItem getCopy( ScriptCopier copier ) {
        return copyInto( new LazyText(), copier );
    }
}
