/*
 * $Id: Text.java,v 1.6 2001/08/10 04:55:25 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.api.text;

import com.iv.flash.parser.*;
import com.iv.flash.util.*;
import com.iv.flash.log.*;
import com.iv.flash.api.*;
import com.iv.flash.api.text.*;
import java.io.*;
import java.util.*;

/**
 * This class represent MM Generator undocumented text tag.
 * <P>
 * It contains vector of {@link TextItem}s.
 *
 * @author Dmitry Skavish
 */
public final class Text extends FlashDef implements TextBlock {

    private boolean withAlpha;              // true if color of change record has alpha
    private Rect bounds;                    // original bounds of the text
    private Rect genBounds;                 // bounds from MM Generator tag
    private Matrix matrix;                  // matrix of the text
    private IVVector items = new IVVector();// vector of TextItems
    private TextLayout myLayout;            // layout of this text

    public Text() {}

    public Text( boolean withAlpha )    {
        this.withAlpha = withAlpha;
    }

    public static Text newText() {
        Text t = new Text(true);
        t.matrix = new Matrix();
        return t;
    }

    public int getTag() {
        if( withAlpha ) return Tag.DEFINETEXT2;
        return Tag.DEFINETEXT;
    }

    // ------------------------------------------------------- //
    //                 TextBlock implementation                //

    /**
     * Layout this text
     */
    public void layout() {
        if( myLayout != null ) return;
        doLayout();
    }

    /**
     * Returns vector of {@link TextRecord}s from this text
     * of specified font.
     *
     * @param font   font of text records to be returned
     * @return text records of specified font
     */
    public IVVector getTextRecords( Font font ) {
        layout();
        return myLayout.getTextRecords( font );
    }

    // ------------------------------------------------------- //

    protected void doLayout() {
        myLayout = new TextLayout(this,genBounds!=null?genBounds:bounds);
        myLayout.layout();
    }

    public void collectDeps( DepsCollector dc ) {
        for( int i=0; i<items.size(); i++ ) {
            TextItem t = (TextItem) items.elementAt(i);
            if( t.font != null ) dc.addDep( t.font );
        }
    }

    public void collectFonts( FontsCollector fc ) {
        for( int i=0; i<items.size(); i++ ) {
            TextItem t = (TextItem) items.elementAt(i);
            if( t.font != null ) fc.addFont( t.font, this );
        }
    }

    public void apply( Context context ) {
        if( myLayout != null ) return;
        super.apply( context );
        for( int i=0; i<items.size(); i++ ) {
            TextItem t = (TextItem) items.elementAt(i);
            t.apply( context );
        }
        // after applying context text bound could be changed, so we need to layout text
        doLayout();
    }


    public boolean isConstant() {
        for( int i=0; i<items.size(); i++ ) {
            TextItem t = (TextItem) items.elementAt(i);
            if( !t.isConstant() ) return false;
        }
        return true;
    }

    public Rect getBounds() {
        // do I have to call layout here ? probably yes
        layout();
        return bounds;
    }

    public void setBounds( Rect bounds ) {
        this.bounds = bounds;
    }

    /**
     * Returns vector of all text items ({@link TextItem}) of this text.
     *
     * @return vector of {@link TextItem}
     */
    public IVVector getTextItems() {
        return items;
    }

    /**
     * Sets new vector of text items ({@link TextItem}) for this text.
     *
     * @param items  new vector of {@link TextItem}
     */
    public void setTextItems( IVVector items ) {
        this.items = items;
    }

    /**
     * Adds new text item to this text
     *
     * @param item   new text item to be added
     * @return added text item
     */
    public TextItem addTextItem( TextItem item ) {
        items.addElement(item);
        return item;
    }

    public void setMatrix( Matrix matrix ) {
        this.matrix = matrix;
    }

    public Matrix getMatrix() {
        return matrix;
    }

    /**
     * Parses any DefineText tag.
     * <P>
     * If current parsed file is not generator template
     * then parse text lazy ({@link LazyText}).
     *
     * @param p         parser to parse with
     * @param withAlpha with alpha or not
     * @return LazyText or Text
     */
    public static FlashDef parse( Parser p, boolean withAlpha ) {
        if( !p.getFile().isTemplate() ) {
            return LazyText.parse( p, withAlpha );
        } else {
            Text text = new Text( withAlpha );
            // get id
            text.setID( p.getUWord() );
            // get bounds and matrix
            text.bounds = Rect.parse(p);
            text.matrix = Matrix.parse(p);

            // do not parse futher, generator text tag will follow
            p.skipLastTag();

            return text;
        }
    }

    /**
     * Parses MM Generator define text tag: 0x2A.
     * <P>
     * Creates vector of TextItems
     */
    public void parseGenText( Parser p ) {
        genBounds = Rect.parse(p);

        TextItem item = new TextItem();
        boolean wasText = false;
        for(;;) {
            int code = p.getUByte();
            if( code == 0 ) {
                items.addElement( item );
                break;
            }
            if( (code&0x80) != 0 ) {
                // control record
                if( wasText ) {
                    // save text item from previous record
                    items.addElement( item );
                    item = (TextItem) item.getCopy(null);
                    item.text = null;
                    wasText = false;
                }
                code = code&0x7f;
                switch( code ) {
                    case 0:   // bold, italic
                        item.style = p.getUByte();
                        break;
                    case 1:   // font id
                        item.font = ((FontDef)p.getDef(p.getUWord())).getFont();
                        break;
                    case 2:   // font height
                        item.height = p.getUWord();
                        break;
                    case 3:   // color with alpha
                        item.color = Color.parseRGBA(p);
                        break;
                    case 4:   // 0 - normal, 1 - superscript, 2 - subscript
                        item.script = p.getUByte();
                        break;
                    case 5:   // kerning
                        item.kerning = p.getWord();
                        break;
                    case 8:   // left, center, right, justify
                        item.align = p.getUByte();
                        break;
                    case 9:   // indent
                        item.indent = p.getUWord();
                        break;
                    case 10:  // left margin
                        item.marginleft = p.getUWord();
                        break;
                    case 11:  // right margin
                        item.marginright = p.getUWord();
                        break;
                    case 12:  // line space
                        item.linesp = p.getUWord();
                        break;
                    default:
                        // there is nothing we can do, so just eat one byte
                        Log.log( "unknownGenText", Util.b2h(code) );
                        p.getUByte();
                        break;
                }
            } else {
                // text record
                char[] t;
                int pos;
                if( item.text != null ) {
                    pos = item.text.length();
                    t = new char[pos+code];
                    item.text.getChars(0, pos, t, 0);
                } else {
                    t = new char[code];
                    pos = 0;
                }
                for( int i=0; i<code; i++ ) {
                    char ch = (char) p.getUWord();
                    t[pos++] = ch;
                }
                item.text = new String(t);
                wasText = true;
            }
        }
    }

    public void write( FlashOutput fob ) {
        int start = fob.getPos();     // save for tag
        fob.skip( 6 );    // 6 - long tag
        fob.writeDefID( this );

        layout();
        bounds.write(fob);
        matrix.write(fob);

        myLayout.write(fob);

        int size = fob.getPos()-start-6;  // 6 - long tag
        fob.writeLongTagAt( Tag.DEFINETEXT2, size, start );
    }

    public void printContent( PrintStream out, String indent ) {
        out.println( indent+"Text: id="+getID() );
        bounds.printContent(out, indent+"    ");
        matrix.printContent(out, indent+"    ");

        TextItem last = new TextItem();
        for( int i=0; i<items.size(); i++ ) {
            TextItem t = (TextItem) items.elementAt(i);
            out.println( indent+"    TextItem: '"+t.text+"'" );
            if( t.font        != last.font        ) out.println( indent+"      font  : '"+t.font.getFontName()+"'" );
            if( t.height      != last.height      ) out.println( indent+"      height: "+t.height );
            if( t.style       != last.style       ) out.println( indent+"      style : "+t.style );
            if( t.align       != last.align       ) out.println( indent+"      align : "+t.align );
            if( t.indent      != last.indent      ) out.println( indent+"      indent: "+t.indent );
            if( t.linesp      != last.linesp      ) out.println( indent+"      linesp: "+t.linesp );
            if( t.kerning     != last.kerning     ) out.println( indent+"      kerning: "+t.kerning );
            if( t.script      != last.script      ) out.println( indent+"      script: "+t.script );
            if( t.marginleft  != last.marginleft  ) out.println( indent+"      margin left: "+t.marginleft );
            if( t.marginright != last.marginright ) out.println( indent+"      margin right : "+t.marginright );
            if( t.color       != last.color       ) t.color.printContent(out, indent+"      ");
            last = t;
        }
    }

    protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
        super.copyInto( item, copier );
        ((Text)item).withAlpha = withAlpha;
        ((Text)item).bounds = (Rect) bounds.getCopy(copier);
        ((Text)item).genBounds = genBounds==null?null:(Rect)genBounds.getCopy(copier);
        ((Text)item).matrix = (Matrix) matrix.getCopy(copier);
        ((Text)item).items = items.getCopy(copier);
        ((Text)item).myLayout = myLayout==null?null:myLayout.getCopy(copier);
        return item;
    }

    public FlashItem getCopy( ScriptCopier copier ) {
        return copyInto( new Text(), copier );
    }
}
