/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.commands;

import com.iv.flash.parser.*;
import com.iv.flash.api.*;
import com.iv.flash.util.*;
import com.iv.flash.log.*;
import java.io.*;

public class GeneralListCommand extends GenericCommand {

    public static final int STARTX = 0;     // 20
    public static final int STARTY = 0;     // 20

    protected String  datasource;
    protected String  halign;
    protected String  valign;
    protected boolean mask;
    protected int     itemspace;
    protected String  instancename;
    protected boolean isVertical;
    protected boolean isFixed;

    protected int listSize; // in twips
    protected int winWidth;
    protected int winHeight;

    protected void initParms( Context context ) throws IVException {
        datasource   = getParameter( context, "datasource", "" );
        halign       = getParameter( context, "halign" );
        valign       = getParameter( context, "valign" );
        mask         = getBoolParameter( context, "mask", true );
        itemspace    = getIntParameter( context, "itemspace", 0 ) * 20;
        instancename = getParameter( context, "instancename" );

        String orient = getParameter( context, "orient", "horizontal" );
        String spacing = getParameter( context, "spacing", "auto" );
        isVertical = orient.equalsIgnoreCase("vertical");
        isFixed    = spacing.equalsIgnoreCase("fixed");
    }

    protected Script makeList( FlashFile file, Context context, Script parent, int frameNum ) throws IVException {
        String[][] data;
        try {
            UrlDataSource ds = new UrlDataSource(datasource,file);
            data = ds.getData();
        } catch( IOException e ) {
            throw new IVException( e, "errDataRead", datasource, getCommandName() );
        }

        if( data.length < 1 ) {
            throw new IVException( "invalDataSource", datasource, getCommandName() );
        }

        Instance mainInst = getInstance();
        Matrix deRotate = deRotateMatrix( mainInst.matrix );
        Rect winBounds = deRotate.transform(new Rect(-1024, -1024, 1024, 1024)); // mask of the list
        winWidth = winBounds.getWidth();
        winHeight = winBounds.getHeight();

        int clipIdx = findColumn( "clip", data );
        int spaceIdx = findColumn( "space", data );
        int instancenameIdx = findColumn( "instancename", data );
        if( clipIdx == -1 ) {
            throw new IVException( "colNotFoundCmd", "Clip", datasource, getCommandName() );
        }

        Script listScript = new Script(1);
        Frame frame = listScript.newFrame();
        int x = STARTX, y = STARTY;

        // process datasource
        for( int row=1; row<data.length; row++ ) {
            Context myContext = makeContext( context, data, row );
            String clipName = data[row][clipIdx];
            Script template = file.getScript(clipName);
            if( template == null ) {
                Log.log( "cmdScriptNotFound", clipName, "List" );
            }
            else {
                Script myScript = template.copyScript();
                file.processScript( myScript, myContext );

                int delta = 0;
                if( spaceIdx != -1 ) {
                    delta = Util.toInt( data[row][spaceIdx], 0 ) * 20;
                }
                Matrix matrix = new Matrix();
                matrix.translate(x,y);
                Instance scInst = frame.addInstance(myScript, row, matrix, null);
                if( instancenameIdx != -1 ) {
                    scInst.name = data[row][instancenameIdx];
                }

                Rect bounds = myScript.getBounds();
                int width = bounds.getWidth();
                int height = bounds.getHeight();
                if( isVertical ) {
                    if( !isFixed ) {
                        delta += height;
                    }
                    delta += itemspace;
                    y += delta;
                } else {
                    if( !isFixed ) {
                        delta += width;
                    }
                    delta += itemspace;
                    x += delta;
                }

                // alignment
                double shiftX, shiftY;
                if( isVertical ) {
                    shiftX = winWidth;
                    shiftY = height;
                } else {
                    shiftX = width;
                    shiftY = winHeight;
                }
                if( halign.equalsIgnoreCase("right") ) {
                    matrix.translateX( shiftX );
                } else if( halign.equalsIgnoreCase("center") ) {
                    matrix.translateX( shiftX/2 );
                }

                if( valign.equalsIgnoreCase("bottom") ) {
                    matrix.translateY( shiftY );
                } else if( valign.equalsIgnoreCase("center") ) {
                    matrix.translateY( shiftY/2 );
                }
            }
        }

        // calculate list size in twips
        if( isVertical ) {
            listSize = y-STARTY;
        } else {
            listSize = x-STARTX;
        }

        Matrix myMatrix = new Matrix();
        myMatrix.translate( -winWidth/2, -winHeight/2 );
        mainInst.matrix = deScaleMatrix( mainInst.matrix );
        mainInst.matrix = mainInst.matrix.multiply(myMatrix);

        if( instancename != null ) {
            mainInst.name = instancename;
        }

        return listScript;
    }

    protected void addMask( Script parent, int frameNum ) {
        if( mask ) {
            // adding mask
            Rect mask_rect = new Rect(0, 0, winWidth, winHeight);
            int maskDepth = parent.reserveLayers( getInstance().depth, 1 );
            Frame maskFrame = parent.getFrameAt(frameNum);
            Instance maskInst = maskFrame.addInstance( getMask(mask_rect), maskDepth, getInstance().matrix, null);
            maskInst.clip = getInstance().depth;
        }
    }
}
