/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.commands;

import com.iv.flash.parser.*;
import com.iv.flash.api.*;
import com.iv.flash.api.shape.*;
import com.iv.flash.api.text.*;
import com.iv.flash.util.*;
import com.iv.flash.log.*;
import com.iv.flash.cache.*;
import java.io.*;
import java.lang.reflect.*;

/**
 * Generic Generator command. All commands have to subclass it.
 */
public class GenericCommand {

    private IVVector names = new IVVector();
    private IVVector values = new IVVector();
    private int type;
    private Instance instance;
    private static final String MACROMEDIACLASS = "com.macromedia.generator.commands.";

    public GenericCommand() {}

    public void doCommand( FlashFile file, Context context, Script parent, int frame )
        throws IVException
    {
    }

    public String getCommandName() {
        String name = getClass().getName();
        int idx = name.lastIndexOf('.');
        return name.substring(idx+1);
    }

    public Instance getInstance() {
        return instance;
    }

    public void setInstance( Instance instance ) {
        this.instance = instance;
    }

    public void printContent( PrintStream out, String indent ) {
        out.println( indent+"Generator Command: "+getClass().getName() );
        out.println( indent+"  Parameters: " );
        for( int i=0; i<values.size(); i++ ) {
            String name = (String) names.elementAt(i);
            String value = (String) values.elementAt(i);
            out.println( indent+"    "+name+"='"+value+"'" );
        }
    }

    public void apply( Context context ) {
        for( int i=0; i<values.size(); i++ ) {
            String value = (String) values.elementAt(i);
            value = context.apply(value);
            values.setElementAt(value, i);
        }
    }

    public void addParameter( String name, String value ) {
        names.addElement( name );
        values.addElement( value );
    }

    public String getParameter( String name ) {
        for( int i=0; i<names.size(); i++ ) {
            String n = (String) names.elementAt(i);
            if( n.equals(name) ) return (String) values.elementAt(i);
        }
        return null;
    }

    public String getParameter( String name, String defaultValue ) {
        String res = getParameter( name );
        if( res != null ) return res;
        return defaultValue;
    }

    public boolean getBoolParameter( String name ) {
        return Util.toBool( getParameter( name ), false );
    }

    public boolean getBoolParameter( String name, boolean def ) {
        return Util.toBool( getParameter( name ), def );
    }

    public String getParameter( Context context, String name, String def ) {
        return context.apply( getParameter( name, def ) );
    }

    public String getParameter( Context context, String name ) {
        return context.apply( getParameter( name ) );
    }

    public boolean getBoolParameter( Context context, String name, boolean def ) {
        String v = context.apply( getParameter( name ) );
        return Util.toBool( v, def );
    }

    public AlphaColor getColorParameter( Context context, String name, AlphaColor def ) {
        String v = context.apply( getParameter( name ) );
        return Util.toColor( v, def );
    }

    public int getIntParameter( Context context, String name, int def ) {
        String v = context.apply( getParameter( name ) );
        return Util.toInt( v, def );
    }

    public double getDoubleParameter( Context context, String name, double def ) {
        String v = context.apply( getParameter( name ) );
        return Util.toDouble( v, def );
    }

    protected Context makeContext( Context context, String[][] data, int row ) throws IVException {
        Context myContext = new StandardContext( context );
        try {
            for( int i=0; i<data[row].length; i++ ) {
                myContext.setValue( data[0][i], data[row][i] );
            }
        } catch( ArrayIndexOutOfBoundsException e ) {
            throw new IVException( e, "invalDataSource", "", getCommandName() );
        }
        return myContext;
    }

    public int findColumn( String columnName, String[][] data ) {
        for( int i=0; i<data[0].length; i++ ) {
            String name = data[0][i];
            if( name.equalsIgnoreCase(columnName) ) return i;
        }
        return -1;
    }

    public Shape getMask( Rect r ) {
        Shape s = new Shape();
        s.setFillStyle0( FillStyle.newSolid( AlphaColor.white ) );
        s.setFillStyle1( 0 );
        s.setLineStyle( new LineStyle(1, AlphaColor.white) );
        s.rectangle(r);
        s.setBounds(r);
        return s;
    }

    /**
     * Discard 'scale' of the matrix<BR>
     * - retrieve scale of the matrix and create new one
     * which is descale of the first and then multiply them
     * in right order
     *
     * @param m      matrix to discard scale from
     * @return same matrix but scale is 1
     */
    public Matrix deScaleMatrix( Matrix m ) {
        final int c1 = 10000;
        Matrix m1 = new Matrix();
        boolean mult = false;
        double sz = getDist( m.transform( new Rect( 0, 0, 0, c1 ) ) );
        if( Math.abs(sz-c1) > 1e-2 ) {
            m1.setScaleY( c1/sz );
            mult = true;
        }
        sz = getDist( m.transform( new Rect( 0, 0, c1, 0 ) ) );
        if( Math.abs(sz-c1) > 1e-2 ) {
            m1.setScaleX( c1/sz );
            mult = true;
        }
        if( mult ) return m.multiply(m1);
        return m;
    }

    /**
     * Discard 'rotate' of the matrix<BR>
     * - retrieve scale and transform of the matrix and create
     * new one from them
     *
     * @param m      matrix to discard rotate from
     * @return same matrix but without rotate
     */
    public Matrix deRotateMatrix( Matrix m ) {
        if( m.getRotateSkew0() == 0.0 && m.getRotateSkew1() == 0.0 ) return m;
        final int c1 = 10000;
        Matrix m1 = new Matrix();
        double sz = getDist( m.transform( new Rect( 0, 0, 0, c1 ) ) );
        if( Math.abs(sz-c1) > 1e-2 ) {
            m1.setScaleY( sz/c1 );
        }
        sz = getDist( m.transform( new Rect( 0, 0, c1, 0 ) ) );
        if( Math.abs(sz-c1) > 1e-2 ) {
            m1.setScaleX( sz/c1 );
        }
        m1.setTranslateX( m.getTranslateX() );
        m1.setTranslateY( m.getTranslateY() );
        return m1;
    }

    /**
     * Return distance between two points defined by the Rect
     *
     * @param r      rectangle which defines two points
     * @return distance
     */
    public double getDist( Rect r ) {
        double dx = r.getWidth();
        double dy = r.getHeight();
        return Math.sqrt( dx*dx + dy*dy );
    }

    public Font getFont( FlashFile file, String fontName ) {
        FlashFile defFile = file.getDefaultSymbolFile();
        if( defFile != null ) {
            return defFile.getFont( fontName );
        }
        return null;
    }

    public Text newText( FlashFile file, String msg, Font font, int size, AlphaColor color, Rect rect ) {
        Text text = Text.newText();
        TextItem item = new TextItem( msg, font, size, color );
        text.addTextItem( item );
        text.setBounds( rect );
        return text;
    }

    public Text newText( FlashFile file, String msg, Font font, int size, AlphaColor color, int width, int height ) {
        return newText( file, msg, font, size, color, new Rect(0,0,width,height) );
    }

    public Text newText( FlashFile file, String msg, String fontName, int size, AlphaColor color, Rect rect ) {
        Font font = getFont( file, fontName );
        if( font == null ) return null;
        return newText( file, msg, font, size, color, rect );
    }

    public void setType( int type ) {
        this.type = type;
    }

    public int getType() {
        return type;
    }

    public static GenericCommand parse( Parser p ) {
        // 0x01 - movie
        // 0x02 - button
        // 0x04 - graphics
        // 0x08 - global
        int flags = p.getUByte();

        p.getUByte(); // skip 1 byte

        int depth = p.getUWord();
        if( (flags&0x04) != 0 ) { // graphics
            // 0x0010 - loop
            // 0x0020 - play once
            // 0x0040 - single frame
            int gflags = p.getUWord();
            int frame = p.getUWord();
        }

        String command = p.getString();
        int clen = command.length();
        int idx = command.indexOf(' ');
        String className;
        if( idx == -1 ) {
            className = command;
            idx = clen;
        } else {
            className = command.substring(0,idx);
        }

        if( className.startsWith(MACROMEDIACLASS) ) {
            className = "com.iv.flash.commands."+
                        className.substring(MACROMEDIACLASS.length());
        }
        Class cmdClazz = CommandCache.getCommandClass(className);
        if( cmdClazz == null ) {
            Log.log( "cmdNotFound", className );
            return null;
        }
        GenericCommand cmd = null;
        try {
            cmd = (GenericCommand) cmdClazz.newInstance();
        } catch( Exception e ) {
            Log.log( e, "cmdNotCreated", className );
            return null;
        }

        cmd.setType( flags );

        /* parse parameters, idx = index of blank
         *
         * format:
         *  commandname name1="value1" name2="value/n/rvalue/n/rvalue\"value\"" name3="value3"
         */
        StringBuffer sb = new StringBuffer();
        for( ; idx+1<clen; ) {
            int startName = idx+1;
            int endName = command.indexOf('=', startName);
            String name = command.substring(startName, endName);
            idx = endName+1;
            sb.setLength(0);
            boolean quotes = false;
        ValueLoop:
            while( idx < clen ) {
                char ch = command.charAt(idx);
                switch( ch ) {
                    case '\\':
                        sb.append( command.charAt(idx+1) );
                        idx += 2;
                        break;
                    case '/':
                        if( clen-idx >= 4 && command.charAt(idx+2)=='/' &&
                            ( (command.charAt(idx+1)=='n' && command.charAt(idx+3)=='r' ) ||
                              (command.charAt(idx+1)=='r' && command.charAt(idx+3)=='n' )
                            )
                          ) {
                            sb.append('\n');
                            idx += 4;
                        } else {
                            sb.append(ch);
                            idx++;
                        }
                        break;
                    case '"':
                        quotes = !quotes;
                        idx++;
                        break;
                    case ' ':
                        if( !quotes ) break ValueLoop;
                    default:
                        sb.append(ch);
                        idx++;
                        break;
                }
            }
            String value = new String(sb);
            cmd.addParameter( name, value );
        }

        return cmd;
    }

    protected GenericCommand copyInto( GenericCommand cmd, Instance instance ) {
        ((GenericCommand)cmd).type = type;
        ((GenericCommand)cmd).instance = instance;
        ((GenericCommand)cmd).names = new IVVector(names);
        ((GenericCommand)cmd).values = new IVVector(values);
        return cmd;
    }

    public GenericCommand getCopy( Instance instance ) {
        try {
            Constructor c = getClass().getConstructor( new Class [] {} );
            GenericCommand o = (GenericCommand) c.newInstance( new Object[] {} );
            return copyInto( o, instance );
        } catch( Exception e ) {
            return null;
        }

    }
}
