/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.commands;

import com.iv.flash.parser.*;
import com.iv.flash.api.*;
import com.iv.flash.util.*;
import com.iv.flash.log.*;

import java.io.*;
import java.util.*;
import javax.xml.transform.*;

import org.apache.xerces.parsers.DOMParser;
import org.apache.xpath.*;
import org.apache.xpath.objects.*;
import org.apache.xml.utils.*;
import org.w3c.dom.*;
import org.w3c.dom.traversal.*;
import org.xml.sax.SAXException;

/**
 * Base class for most XML commands
 */
public class GenericXMLCommand extends GenericCommand {

    protected String datasource;
    protected String select;

    /**
     * Init common parameters of xml based commands
     */
    protected void initParms( Context context ) throws IVException {
        datasource   = getParameter( context, "datasource" );
        select       = getParameter( context, "select" );
    }

    /**
     * Retrieve nearest parent XMLContext from contexts' stack
     *
     * @param context context to start searching from
     * @return found xmlcontext or null
     */
    protected XMLContext retrieveXMLContext( Context context ) {
        while( context != null && !(context instanceof XMLContext) ) {
            context = context.getParent();
        }
        return (XMLContext) context;
    }

    /**
     * Return XMLContext for this command, either by creating one from datasource parameter
     * or returning nearest parent XMLContext or null
     *
     * @param file current FlashFile
     * @param context current context for this command
     * @return XMLContext for the command or null
     */
    protected XMLContext getXMLContext( FlashFile file, Context context ) throws IVException {
        if( datasource != null ) {
            Node node = null;
            try {
                XMLDataSource ds = new XMLDataSource(datasource,file);
                node = ds.getNode();
            } catch( Exception e ) {
                throw new IVException( e, "errDataRead", datasource, getCommandName() );
            }
            return new XMLContext(context, node);
        }
        return retrieveXMLContext( context );
    }

    /**
     * Select set of nodes from given XMLContext
     *
     * @param context given XMLContext to select from
     * @return list of selected nodes (may be empty)
     */
    protected NodeList select( XMLContext context ) throws IVException {
        return select(context,select);
    }

    /**
     * Select set of nodes from given XMLContext
     *
     * @param context given XMLContext to select from
     * @return list of selected nodes (may be empty)
     */
    protected NodeList select( XMLContext context, String select ) throws IVException {
        if( select == null ) {
            select = "*";   // get all children
        }
        try {
            return context.selectNodeList( select );
        } catch( TransformerException e ) {
            throw new IVException(e);
        }
    }

    /**
     * Convert NodeList to ArrayList
     */
    protected ArrayList toArrayList( NodeList nl ) {
        int len = nl.getLength();
        ArrayList al = new ArrayList(len);
        for( int i=0; i<len; i++ ) {
            al.add( nl.item(i) );
        }
        return al;
    }

    /**
     * Sort given NodeList using given XPath 'sortby' as a key
     *
     * @param nl NodeList to sort (sort in place)
     * @param sortby XPath expression to sort on
     * @param ascending if true then sort in ascending order, otherwise in descending
     * @param sort do not sort if false, just convert NodeList to ArrayList
     * @return sorted ArrayList of SortListNodes
     */
    protected ArrayList sort( NodeList nl, XPath sortby, final boolean ascending, boolean sort ) {
        int len = nl.getLength();
        // create array to sort
        ArrayList items = new ArrayList(len);

        XPathContext xpath_context = new XPathContext();
        // evaluate keys
        for( int i=0; i<len; i++ ) {
            Node node = (Node) nl.item(i);
            if( sort ) {
                String key;
                try {
                    XObject xo = XMLHelper.evalXPath(xpath_context, node, sortby);
                    key = XMLHelper.getXObjectData(xo);
                } catch( TransformerException e ) {
                    Log.log(e);
                    key = "";
                }
                items.add( new SortListNode(node,key) );
            } else {
                items.add( new SortListNode(node,"") );
            }
        }

        if( !sort ) return items;

        // sort
        Collections.sort( items, new java.util.Comparator() {
                public int compare( Object o1, Object o2 ) {
                    SortListNode i1 = (SortListNode) o1;
                    SortListNode i2 = (SortListNode) o2;
                    if( ascending ) {
                        return i1.key.compareTo( i2.key );
                    } else {
                        return i2.key.compareTo( i1.key );
                    }
                }
            }
        );

        return items;
    }

    /**
     * Get command's parameter by name, substitute possible variables using given context
     * and compile it to XPath
     *
     * @param context context to apply for parameter
     * @param name name of the parameter
     * @param def default value
     * @return compiled XPath or null
     */
    protected XPath getXPathParameter( Context context, String name, String def ) {
        try {
            String expr = getParameter( context, name );
            if( expr == null ) expr = def;
            if( expr == null ) return null;
            return XMLHelper.getXPath(expr);
        } catch( Exception e ) {
            Log.log(e);
            try {
                if( def == null ) return null;
                return XMLHelper.getXPath(def);
            } catch( Exception ee ) {
                Log.log(ee);
                return null;
            }
        }
    }

    /**
     * Evaluate given XPath in given XMLContext and return the result as a String
     *
     * @param context XMLContext to evaluate XPath in
     * @param xpath XPath which will be executed
     * @param def default value returned in case of any errors
     * @return result as a string
     */
    protected String evalStringParameter( XMLContext context, XPath xpath, String def ) {
        try {
            XObject xo = context.eval(xpath);
            return xo.toString();
        } catch( Exception e ) {
            Log.log(e);
            return def;
        }
    }

    /**
     * Evaluate given XPath in given XMLContext and return the result as a boolean
     *
     * @param context XMLContext to evaluate XPath in
     * @param xpath XPath which will be executed
     * @param def default value returned in case of any errors
     * @return result as boolean
     */
    protected boolean evalBoolParameter( XMLContext context, XPath xpath, boolean def ) {
        try {
            XObject xo = context.eval(xpath);
            if( xo.getType() == XObject.CLASS_BOOLEAN ) return xo.bool();
            return def;
        } catch( Exception e ) {
            Log.log(e);
            return def;
        }
    }

    /**
     * Evaluate given XPath in given XMLContext and return the result as a int
     *
     * @param context XMLContext to evaluate XPath in
     * @param xpath XPath which will be executed
     * @param def default value returned in case of any errors
     * @return result as a int
     */
    protected int evalIntParameter( XMLContext context, XPath xpath, int def ) {
        return (int) evalDoubleParameter( context, xpath, def );
    }

    /**
     * Evaluate given XPath in given XMLContext and return the result as a double
     *
     * @param context XMLContext to evaluate XPath in
     * @param xpath XPath which will be executed
     * @param def default value returned in case of any errors
     * @return result as a double
     */
    protected double evalDoubleParameter( XMLContext context, XPath xpath, double def ) {
        try {
            XObject xo = context.eval(xpath);
            if( xo.getType() == XObject.CLASS_NUMBER ) return xo.num();
            return def;
        } catch( Exception e ) {
            Log.log(e);
            return def;
        }
    }

}

