/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.commands;

import com.iv.flash.parser.*;
import com.iv.flash.api.*;
import com.iv.flash.api.image.*;
import com.iv.flash.api.shape.*;
import com.iv.flash.util.*;
import com.iv.flash.cache.*;
import com.iv.flash.url.*;
import com.iv.flash.log.*;
import java.io.*;

/**
 * Insert Media generator command<BR>
 * Insert gif, jpg, png, swf or swt<BR>
 * Detect media type from its data, not from extension<BR>
 *
 * <P>if 'scalex' or 'scaley' are not null, then scale width or/and height
 * of the image accordingly, the placeholder's scaling is ignored
 * <P>if both 'scalex' and 'scaley' are null then check 'scale'. if it's true
 * then scale image to the placeholder size, if scale is false then leave the
 * image as is and ignore the placeholder's scaling
 *
 * @author Dmitry Skavish
 */
public class InsertMediaCommand extends GenericCommand {

    public InsertMediaCommand() {}

    public void doCommand( FlashFile file, Context context, Script parent, int frame ) throws IVException {
        String filename = getParameter( context, "filename", "" );
        boolean cache   = getBoolParameter( context, "cache", false );
        boolean scale   = getBoolParameter( context, "scale", false );
        boolean expand  = getBoolParameter( context, "expand", false );
        int scaletox    = getIntParameter( context, "scalex", -1 ) * 20;
        int scaletoy    = getIntParameter( context, "scaley", -1 ) * 20;
        String halign   = getParameter( context, "halign", "left" );
        String valign   = getParameter( context, "valign", "top" );
        boolean mask    = getBoolParameter( context, "mask", false );
        //    String export = context.apply( getParameter("export","JPEG") );
        //    String quality = context.apply( getParameter("quality","80") );
        String instancename = getParameter( context, "instancename" );

        Instance inst = getInstance();

        IVUrl url = IVUrl.newUrl(filename, file);
        filename = url.getName();

        Object media = null;

        // check in cache first
        if( cache ) {
            media = MediaCache.getMedia( filename );
        }

        boolean inExternals = false;
        // check may be it's in external files
        if( media == null ) {
            media = file.getExternalFile( filename );
            inExternals = media != null;
        }

        // read from url
        if( media == null ) {
            InputStream is = null;
            FlashBuffer fb = null;
            try {
                is = url.getInputStream();
                fb = new FlashBuffer(is);
            } catch( IOException e ) {
                throw new IVException( e, "errCmdFileRead", filename, getCommandName() );
            } finally {
                try {
                    if( is != null ) is.close();
                } catch( IOException e ) {}
            }

            // detect media type
            if( fb.getUByte() == 'F' ) {    // check for swf format
                if( fb.getUByte() == 'W' && fb.getUByte() == 'S' ) {
                    media = FlashFile.parse(filename, fb);
                    file.addExternalFile(filename, (FlashFile) media);
                    inExternals = true;
                } else {
                    throw new IVException( "unsupMedia", filename );
                }
            } else {
                media = Bitmap.newBitmap(fb);
                if( media == null ) {
                    throw new IVException( "unsupMedia", filename );
                }
            }
            MediaCache.addMedia(url, media, fb.getSize(), cache);
        }

        Script script;
        int width, height;

        if( media instanceof FlashFile ) {
            // insert flash movie
            FlashFile flashFile = (FlashFile) media;
            if( !inExternals ) file.addExternalFile(filename, flashFile);
            script = flashFile.getMainScript();
            script.resetMain();
            if( flashFile.isTemplate() ) {
                script = script.copyScript();
                file.processScript(script, context);
            }

            Rect r = flashFile.getFrameSize();  // script.getBounds();
            width = r.getWidth();   // in twixels
            height = r.getHeight(); // in twixels

            // expand frames
            if( expand ) {
                int total = parent.getFrameCount();
                int myTotal = script.getFrameCount();
                if( total-frame < myTotal ) {
                    int numberToAdd = (frame+myTotal) - total;
                    for( int i=0; i<numberToAdd; i++ ) parent.newFrame();
                }
            }

        } else {
            // insert image file
            Bitmap bitmap = (Bitmap) media;
            Instance myInst = bitmap.newInstance();
            width = bitmap.getWidth()*20;   // now in twixels
            height = bitmap.getHeight()*20; // now in twixels

            script = inst.copyScript();
            Frame myFrame = script.newFrame();
            myFrame.addInstance( myInst, 1 );
        }

        inst.setScript( script );

        Matrix myMatrix = new Matrix();
        if( scaletox >= 0 || scaletoy >= 0 ) {
            if( scaletox >= 0 ) {
                myMatrix.setScaleX( (double) scaletox / width );
                width = scaletox;
            }
            if( scaletoy >= 0 ) {
                myMatrix.setScaleY( (double) scaletoy / height );
                height = scaletoy;
            }
            scale = false;  // just to make sure
        } else if( scale ) {
            myMatrix.setScaleX( 2048.0/width );
            myMatrix.setScaleY( 2048.0/height );
        }

        if( !scale ) {
            // remove rotation from instance's matrix
            Matrix deRotate = deRotateMatrix( inst.matrix );
            // placeholder's bounds
            Rect winBounds = deRotate.transform(new Rect(-1024, -1024, 1024, 1024));
            int winWidth = winBounds.getWidth();
            int winHeight = winBounds.getHeight();

            // horizontal alignment
            if( halign.equalsIgnoreCase( "left" ) ) {
                myMatrix.translateX( -winWidth/2 );
            } else if( halign.equalsIgnoreCase( "right" ) ) {
                myMatrix.translateX( -width+winWidth/2 );
            } else {    // center
                myMatrix.translateX( -width/2 );
            }

            // vertical alignment
            if( valign.equalsIgnoreCase( "top" ) ) {
                myMatrix.translateY( -winHeight/2 );
            } else if( valign.equalsIgnoreCase( "bottom" ) ) {
                myMatrix.translateY( -height+winHeight/2 );
            } else {    // center
                myMatrix.translateY( -height/2 );
            }

            inst.matrix = deScaleMatrix( inst.matrix );
        } else {
            myMatrix.setTranslateX( -1024 );    // -(myMatrix.getScaleX()*width)/2
            myMatrix.setTranslateY( -1024 );    // -(myMatrix.getScaleY()*height)/2
        }

        inst.matrix = inst.matrix.multiply(myMatrix);

        if( instancename != null ) {
            inst.name = instancename;
        }
    }

}
