/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.commands;

import com.iv.flash.parser.*;
import com.iv.flash.api.*;
import com.iv.flash.api.image.*;
import com.iv.flash.api.shape.*;
import com.iv.flash.api.text.*;
import com.iv.flash.util.*;
import com.iv.flash.cache.*;
import com.iv.flash.url.*;
import com.iv.flash.log.*;
import java.io.*;
import java.util.*;

/**
 * Insert Text generator command<BR>
 *
 * @author Dmitry Skavish
 */
public class InsertTextCommand extends GenericCommand {

    public InsertTextCommand() {}

    public void doCommand( FlashFile file, Context context, Script parent, int frame ) throws IVException {
        String filename  = getParameter( context, "filename" );
        String text      = getParameter( context, "text" );
        boolean mask     = getBoolParameter( context, "mask", false );          // true, false
        boolean html     = getBoolParameter( context, "html", false );          // true, false
        String embedset  = getParameter( context, "embedset", "[all]" );
        AlphaColor color = getColorParameter( context, "fontColor", AlphaColor.red );
        int fontSize     = getIntParameter( context, "fontSize", 12 ) * 20;
        String fontType  = getParameter( context, "fontType", "Arial" );
        String fftFile   = getParameter( context, "fftFilePath" );              // name of fft file
        boolean bold     = getBoolParameter( context, "bold", false );          // true, false
        boolean italic   = getBoolParameter( context, "italic", false );        // true, false
        String alignment = getParameter( context, "alignment", "left" );        // left, right, center, justify
        double spacing   = getDoubleParameter( context, "letterSpacing", 0.0 ) * 10; // in half pixels
        int lineSpacing  = getIntParameter( context, "lineSpacing", 0 ) * 20;   // in points
        int transparency = getIntParameter( context, "transparency", 100 );     // 0-100
        boolean cache   = getBoolParameter( context, "cache", false );
        String instancename = getParameter( context, "instancename" );

        Instance inst = getInstance();

        // -----------------------------------------------------------
        // read text from file if needed
        // -----------------------------------------------------------
        if( text == null ) {
            if( filename == null ) {
                throw new IVException( "noText" );
            }
            IVUrl url = IVUrl.newUrl(filename, file);

            // check in cache first
            if( cache ) {
                text = (String) MediaCache.getMedia( url );
            }

            if( text == null ) {
                InputStream is = null;
                try {
                    is = url.getInputStream();
                    BufferedReader reader = new BufferedReader( new InputStreamReader(is) );
                    StringBuffer sb = new StringBuffer(100);
                    String s = reader.readLine();
                    while( s != null ) {
                        sb.append( s );
                        sb.append( '\n' );
                        s = reader.readLine();
                    }
                    text = sb.toString();
                } catch( IOException e ) {
                    throw new IVException( e, "errCmdFileRead", url.getName(), getCommandName() );
                } finally {
                    try {
                        if( is != null ) is.close();
                    } catch( IOException e ) {}
                }
                MediaCache.addMedia( url, text, text.length(), cache );
            }
        }

        // ---------------------------------------------------------------
        // get the font (it's a little bit ugly, needs to be changed!!!)
        // ---------------------------------------------------------------
        if( fftFile == null ) {
            fftFile = fontType.replace( ' ', 'G' )+".fft";
            if( italic ) fftFile = "I"+fftFile;
            if( bold ) fftFile = "B"+fftFile;
        }

        fftFile = Util.translatePath( fftFile );
        File fontFile = new File( fftFile );
        if( !fontFile.exists() ) {
            fftFile = Util.concatFileNames( PropertyManager.fontPath, fftFile );
            fontFile = new File( Util.installDir, fftFile );
        }
        fftFile = fontFile.getAbsolutePath();

        Font font = FontDef.load( fftFile, file );

        // create text block and set parameters
        Text t_block = Text.newText();
        color.setAlpha( (transparency * 255)/100 );
        TextItem t_item = new TextItem( text, font, fontSize, color );

        if( alignment.equals("left") ) t_item.align = 0;
        else if( alignment.equals("right") ) t_item.align = 1;
        else if( alignment.equals("center") ) t_item.align = 2;
        else if( alignment.equals("justify") ) t_item.align = 3;

        t_item.kerning = (int) spacing;
        t_item.linesp = lineSpacing;

        t_block.addTextItem( t_item );

        // calculate text bounds
        Matrix deRotate = deRotateMatrix( inst.matrix );
        Rect winBounds = deRotate.transform(new Rect(-1024, -1024, 1024, 1024)); // mask of the list
        int winWidth = winBounds.getWidth();
        int winHeight = winBounds.getHeight();
        t_block.setBounds( new Rect(0,0,winWidth,winHeight) );

        Matrix myMatrix = new Matrix();
        myMatrix.translate( -winWidth/2, -winHeight/2 );
        inst.matrix = deScaleMatrix( inst.matrix );
        inst.matrix = inst.matrix.multiply(myMatrix);

        // copy script and add the text there
        Script script = inst.copyScript();
        Frame textFrame = script.newFrame();
        textFrame.addInstance( t_block, 1, new Matrix(), null );

        // add mask
        if( mask ) {
            Rect mask_rect = new Rect(0, 0, winWidth, winHeight);
            int maskDepth = parent.reserveLayers( inst.depth, 1 );
            Frame maskFrame = parent.getFrameAt(frame);
            Instance maskInst = maskFrame.addInstance( getMask(mask_rect), maskDepth, inst.matrix, null);
            maskInst.clip = inst.depth;
        }

        if( instancename != null ) {
            inst.name = instancename;
        }
    }

}

