/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.commands;

import com.iv.flash.parser.*;
import com.iv.flash.api.*;
import com.iv.flash.api.shape.*;
import com.iv.flash.api.text.*;
import com.iv.flash.util.*;
import com.iv.flash.log.*;
import java.io.*;
import java.text.*;
import java.util.*;

/*
 * 03/23/2001 fixed some problems with help from Patrick Talbot
 */

/**
 * Insert Table command
 *
 */
public class TableCommand extends GenericCommand {

    public static final int FIXED  = 0;
    public static final int AUTO   = 1;
    public static final int HALF   = 2;
    public static final int DOUBLE = 3;

    protected String  datasource;       // datasource
    protected String  halign;           // left, center, right
    protected String  valign;           // top, center, bottom
    protected int     rows, cols;       // number of rows and columns
    protected int     sizing;           // auto, fixed, half, double
    protected String  defsym;           // default symbol name
    protected String  labelformat;      // Arial, Times, Courier, name of the font
    protected int     labelsizing;      // auto, fixed, half, double
    protected int     labelsize;        // size of label text
    protected boolean borders;          // true, false
    protected AlphaColor bordercolor;   // border color
    protected int     borderthickness;  // thickness of border in twips
    protected String  mediafile;        // external symbol file
    protected String  instancename;     // instancename
    protected String[] rlabels;         // labels for rows
    protected String[] clabels;         // labels for columns

    protected int tableSize; // in twips
    protected Rect winBounds;
    protected int cellWidth, cellHeight;

    public TableCommand() {}

    protected String[] parseLabels( String s ) {
        if( s == null || s.length() == 0 ) return null;
        StringTokenizer st = new StringTokenizer(s, ",\t\n \r");
        IVVector v = new IVVector();
        while( st.hasMoreTokens() ) {
            v.addElement( st.nextToken() );
        }
        String[] rs = new String[v.size()];
        v.copyInto(rs);
        return rs;
    }

    protected int parseLabelSizing( String s ) {
        try {
            labelsize = Integer.parseInt(s) * 20;
            return FIXED;
        } catch( NumberFormatException e ) {
            labelsize = 18 * 20;
            return parseSizing( s );
        }
    }

    protected int parseSizing( String s ) {
        if( s.equalsIgnoreCase("fixed")  ) return FIXED;
        if( s.equalsIgnoreCase("auto")   ) return AUTO;
        if( s.equalsIgnoreCase("half")   ) return HALF;
        if( s.equalsIgnoreCase("double") ) return DOUBLE;
        return AUTO;
    }

    protected void initParms( Context context ) throws IVException {
        datasource      = getParameter( context, "datasource", "" );
        halign          = getParameter( context, "halign" );
        valign          = getParameter( context, "valign" );
        rows            = getIntParameter( context, "rows", 0 );
        cols            = getIntParameter( context, "cols", 0 );
        sizing          = parseSizing( getParameter( context, "sizing" ) );
        defsym          = getParameter( context, "defsym" );
        labelformat     = getParameter( context, "labelformat" );
        labelsizing     = parseLabelSizing( getParameter( context, "labelsizing" ) );
        rlabels         = parseLabels( getParameter( context, "rlabels" ) );
        clabels         = parseLabels( getParameter( context, "clabels" ) );
        mediafile       = getParameter( context, "mediafile" );
        borders         = getBoolParameter( context, "borders", true );
        bordercolor     = getColorParameter( context, "bordercolor", AlphaColor.black );
        borderthickness = borders? getIntParameter( context, "borderthickness", 20 ): 0;
        instancename = getParameter( context, "instancename" );
    }

    protected Instance putSymbolInCell( Frame frame, FlashDef sym, Rect rect, int layer, int sizing, boolean prop ) {
        Matrix m = new Matrix();
        m.translate(rect.getXMin(), rect.getYMin());
        switch( sizing ) {
            case FIXED:
                break;
            case AUTO: {
                Rect b = sym.getBounds();
                double xk = (double)cellWidth/b.getWidth();
                double yk = (double)cellHeight/b.getHeight();
                if( prop ) {
                    double k = Math.min(xk,yk);
                    m.scale(k);
                } else {
                    m.scale( xk, yk );
                }
                break;
            }
            case HALF:
                m.scale( 0.5 );
                break;
            case DOUBLE:
                m.scale( 2.0 );
                break;
        }
        return frame.addInstance(sym, layer, m, null);
    }

    protected Rect getCellRect( int col, int row ) {
        Rect r = new Rect();
        r.setXMin( borderthickness+(cellWidth+borderthickness)*col );
        r.setYMin( borderthickness+(cellHeight+borderthickness)*row );
        r.setXMax( r.getXMin()+cellWidth );
        r.setYMax( r.getYMin()+cellHeight );
        return r;
    }

    protected Script makeTable( FlashFile file, Context context, Script parent, int frameNum ) throws IVException {

        String[][] data;
        try {
            UrlDataSource ds = new UrlDataSource(datasource,file);
            data = ds.getData();
        } catch( IOException e ) {
            throw new IVException( e, "errDataRead", datasource, getCommandName() );
        }

        if( data.length < 1 ) {
            throw new IVException( "invalDataSource", datasource, getCommandName() );
        }

        winBounds = new Rect(-1024, -1024, 1024, 1024);
        Instance mainInst = getInstance();
        winBounds = mainInst.matrix.transform(winBounds); // main window
        int winWidth = winBounds.getWidth();
        int winHeight = winBounds.getHeight();

        int clipIdx = findColumn( "clip", data );

        if( rows <= 0 || cols <= 0 ) {
            throw new IVException( "rowsOrCols", getCommandName() );
        }

        boolean isLabels = rlabels != null || clabels != null;
        if( defsym != null || isLabels ) {
            // loading symbol file
            if( mediafile != null ) {
                try {
                    file.addExternalFile( mediafile, true );
                } catch( IVException e ) {
                    Log.log( e );
                }
            }
            // load default symbol file
            file.getDefaultSymbolFile();
        }


        Script tableScript = new Script(1);
        Frame frame = tableScript.newFrame();

        int totalRows = rows;
        int totalCols = cols;
        if( rlabels != null ) totalCols++;
        if( clabels != null ) totalRows++;

        // calculate cell width and height
        cellWidth  = (winWidth-borderthickness*(totalCols+1)) / totalCols;
        cellHeight = (winHeight-borderthickness*(totalRows+1)) / totalRows;

        if( cellWidth <= 0 || cellHeight <= 0 ) {
            Log.log( "borderTooThick" );
        }

        // buidl grid cells
        Rect[][] cellRects = new Rect[totalCols][totalRows];
        for( int c=0; c<totalCols; c++ ) {
            for( int r=0; r<totalRows; r++ ) {
                cellRects[c][r] = getCellRect( c, r );
            }
        }
        // draw grid (borders)
        if( borders ) {
            Shape shape = new Shape();
            int fullColor = shape.addLineStyle( new LineStyle(25, bordercolor) );
            int reducedColor = shape.addLineStyle( new LineStyle(25, bordercolor.getReducedColor()) );
            // draw cells with full color
            shape.setLineStyle(fullColor);
            for( int c=0; c<totalCols; c++ ) {
                for( int r=0; r<totalRows; r++ ) {
                    Rect rect = cellRects[c][r];
                    shape.moveTo(rect.getXMin(), rect.getYMax());
                    shape.lineTo(rect.getXMin(), rect.getYMin());
                    shape.lineTo(rect.getXMax(), rect.getYMin());
                }
            }
            // draw main table with full color
            shape.moveTo(winWidth, 0);
            shape.lineTo(winWidth, winHeight);
            shape.lineTo(0, winHeight);

            // draw cells with reduced color
            shape.setLineStyle(reducedColor);
            for( int c=0; c<totalCols; c++ ) {
                for( int r=0; r<totalRows; r++ ) {
                    Rect rect = cellRects[c][r];
                    shape.moveTo(rect.getXMax(), rect.getYMin());
                    shape.lineTo(rect.getXMax(), rect.getYMax());
                    shape.lineTo(rect.getXMin(), rect.getYMax());
                }
            }
            // draw main table with reduced color
            shape.moveTo(0, winHeight);
            shape.lineTo(0, 0);
            shape.lineTo(winWidth, 0);

            // add shape to frame
            frame.addInstance(shape, 2, new Matrix(), null);
            shape.setBounds( new Rect(0, 0, winWidth, winHeight) );
        }

        Script defSymbol = null;
        if( defsym != null ) {
            defSymbol = file.getScript(defsym);
        }

        Font labelFont = null;
        if( isLabels ) {
            if( labelformat == null ) labelformat = "Arial";
            // map generator font names to font names from DefaultSymbolFile
            if( labelformat.equals( "Times" ) ) labelformat = "Times New Roman";
            else if( labelformat.equals( "Courier" ) ) labelformat = "Courier New";

            labelFont = getFont( file, labelformat );
        }

        if( clipIdx == -1 && defSymbol == null ) {
            throw new IVException( "clipOrDefsym" );
        }

        int col = 0, row = 0;
        int layer = 3;
        int rbase = clabels != null? 1: 0;
        int cbase = rlabels != null? 1: 0;

        // draw labels
        if( labelFont != null ) {
            if( rlabels != null ) {
                int nLabels = Math.min(totalRows, rlabels.length);
                for( int i=0; i<nLabels; i++ ) {
                    String label = rlabels[i];
                    Rect r = cellRects[0][i+rbase];
                    Text text = newText( file, label, labelFont, labelsize, AlphaColor.black, r.getWidth(), r.getHeight() );
                    ((TextItem)text.getTextItems().elementAt(0)).align = 2; // center
                    putSymbolInCell( frame, text, r, layer++, labelsizing, true );
                }
                col++;
            }
            if( clabels != null ) {
                int nLabels = Math.min(totalCols, clabels.length);
                for( int i=0; i<nLabels; i++ ) {
                    String label = clabels[i];
                    Rect r = cellRects[i+cbase][0];
                    Text text = newText( file, label, labelFont, labelsize, AlphaColor.black, r.getWidth(), r.getHeight() );
                    ((TextItem)text.getTextItems().elementAt(0)).align = 2; // center
                    putSymbolInCell( frame, text, r, layer++, labelsizing, true );
                }
                row++;
            }
        }

        // process datasource
        for( int r=1; r<data.length; r++ ) {
            Context myContext = makeContext( context, data, r );
            Script template;
            if( defSymbol != null ) {
                template = defSymbol;
            }
            else {
                String clipName = data[r][clipIdx];
                template = file.getScript(clipName);
                if( template == null ) {
                    throw new IVException( "cmdScriptNotFound", clipName, getCommandName() );
                }
            }
            Script myScript = template.copyScript();
            file.processScript( myScript, myContext );

            Instance myInst = putSymbolInCell( frame, myScript, cellRects[col][row], layer++, sizing, true );

            Matrix matrix = myInst.matrix;
            if( halign.equalsIgnoreCase("right") ) {
                matrix.translateX( cellWidth );
            } else if( halign.equalsIgnoreCase("center") ) {
                matrix.translateX( cellWidth/2 );
            }

            if( valign.equalsIgnoreCase("bottom") ) {
                matrix.translateY( cellHeight );
            } else if( valign.equalsIgnoreCase("center") ) {
                matrix.translateY( cellHeight/2 );
            }

            col++;
            if( col >= totalCols ) {
                row++;
                col = cbase;
                if( row >= totalRows ) break;
            }

        }

        return tableScript;
    }

    protected void addMask( Script parent, int frameNum ) {
        // adding mask
        int maskDepth = parent.reserveLayers( getInstance().depth, 1 );
        Frame maskFrame = parent.getFrameAt(frameNum);
        Instance maskInst = maskFrame.addInstance( getMask(winBounds), maskDepth, new Matrix(), null);
        maskInst.clip = getInstance().depth;
    }

    public void doCommand( FlashFile file, Context context, Script parent, int frameNum ) throws IVException {
        initParms( context );

        Script tableScript = makeTable( file, context, parent, frameNum );

        Matrix mainMatrix = new Matrix();
        mainMatrix.setTranslateX( winBounds.getXMin() );
        mainMatrix.setTranslateY( winBounds.getYMin() );
        getInstance().matrix = mainMatrix;

        if( instancename != null ) {
            getInstance().name = instancename;
        }

        getInstance().setScript( tableScript );
        addMask( parent, frameNum );
    }

}
