/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.log;

import com.iv.flash.api.*;
import com.iv.flash.util.*;
import java.io.*;
import java.net.*;
import java.util.*;

public class Log {

    public static class Message {
        public String key;
        public int verboseLevel;
        public String message;
    }

    public static final int VERBOSE_HIGH   = 3;
    public static final int VERBOSE_MEDIUM = 2;
    public static final int VERBOSE_LOW    = 1;
    public static final int VERBOSE_NONE   = 0;

    private static Hashtable messages = new Hashtable();
    private static int verboseLevel = VERBOSE_HIGH;
    private static Logger logger;
    private static PrintWriter err = new PrintWriter( System.err, true );
    private static String logFileName;

    public static void setLogToConsole() {
        logger = new ConsoleLogger();
    }

    public static void setLogToFile( boolean commit ) {
        String fileName = PropertyManager.getProperty( "com.iv.flash.logFile", "logs/generator.log" );
        logger = new FileLogger(fileName, commit);
    }

    public static void setLogger( Logger customLogger ) {
        logger = customLogger;
    }

    public static String log( String key ) {
        return _log( getMessage( key, null ) );
    }

    public static String log( String key, Object p0 ) {
        return _log( getMessage( key, new Object[] { p0 } ) );
    }

    public static String log( String key, Object p0, Object p1 ) {
        return _log( getMessage( key, new Object[] { p0, p1 } ) );
    }

    public static String log( String key, Object p0, Object p1, Object p2 ) {
        return _log( getMessage( key, new Object[] { p0, p1, p2 } ) );
    }

    public static String log( String key, Object[] p ) {
        return _log( getMessage( key, p ) );
    }

    public static String log( IVException e ) {
        String key = e.getMessageKey();
        Object[] parms = e.getParameters();
        String msg = log( key, parms );
        Exception nested = e.getNestedException();
        if( nested != null ) msg += log( nested );
        return msg;
    }

    public static String log( Exception e ) {
        if( e instanceof IVException ) return log( (IVException) e );
        ByteArrayOutputStream bas = new ByteArrayOutputStream();
        PrintWriter pw = new PrintWriter( bas, true );
        pw.println( e.getLocalizedMessage() );
        if( verboseLevel >= VERBOSE_HIGH ) {
            e.printStackTrace( pw );
        }
        pw.close();
        return _log( bas.toString() );
    }

    public static String log( Exception e, String key ) {
        return log( key ) + log( e );
    }

    public static String log( Exception e, String key, Object p0 ) {
        return log( key, p0 ) + log( e );
    }

    public static String log( Exception e, String key, Object p0, Object p1 ) {
        return log( key, p0, p1 ) + log( e );
    }

    public static String log( Exception e, String key, Object p0, Object p1, Object p2 ) {
        return log( key, p0, p1, p2 ) + log( e );
    }

    public static String log( Exception e, String key, Object[] p ) {
        return log( key, p ) + log( e );
    }

    public static void err( String s ) {
//        System.err.println( s );
    }

    public static void err( Exception e ) {
//        System.err.println( e.getLocalizedMessage() );
//        e.printStackTrace(System.err);
    }

    public static void err( String msg, Exception e ) {
//        err( msg );
//        err( e );
    }

    /* ------------------------------------------------------------------------------- */

    public static String getMessage( String key, Object[] parms ) {
        if( key == null ) return null;
        Message m = (Message) messages.get(key);
        if( m == null ) {
            return "Log Error: Message with key '"+key+"' not found";
        }
        if( m.verboseLevel > verboseLevel ) return null;
        if( parms == null ) return m.message;

        String msg = m.message;
        StringBuffer sb = new StringBuffer( msg.length() );
        for( int i=0; i<msg.length(); i++ ) {
            char ch = msg.charAt(i);
            if( ch == '&' ) {
                i++;
                ch = msg.charAt(i);
                if( ch == '&' ) {
                    sb.append(ch);
                } else {
                    int idx = ch-'0';
                    if( idx < 0 || idx >= parms.length ) {
                        sb.append("<error>");
                    } else {
                        if( parms[idx] != null ) {
                            sb.append( parms[idx].toString() );
                        }
                    }
                }
            } else {
                sb.append(ch);
            }
        }
        return new String(sb);
    }

    private static String _log( String msg ) {
        if( msg == null || msg.length() == 0 ) return "";
        logger.log(msg);
        return msg + Util.lineSeparator;
    }

    private static void loadMessages() {
        try {
            String fileName = PropertyManager.getProperty("com.iv.flash.messageFile","bin/messages.txt");
            File file = Util.getSysFile( fileName );
            BufferedReader reader = new BufferedReader( new FileReader(file) );
            String line;
            int lineNum = 0;
            while( (line=reader.readLine()) != null ) {
                lineNum++;
                int idx = line.indexOf(',');
                if( idx < 0 ) {
                    err( "Error in message file, line "+lineNum );
                    continue;
                }
                String key = line.substring(0, idx).trim();
                int si = idx+1;
                idx = line.indexOf(',',si);
                if( idx < 0 ) {
                    err( "Error in message file, line "+lineNum );
                    continue;
                }
                String verbose = line.substring(si, idx).trim();
                int verboseLevel = Integer.valueOf(verbose).intValue();
                si = idx+1;
                String message = line.substring(si).trim();
                Message m = new Message();
                m.key = key;
                m.verboseLevel = verboseLevel;
                m.message = message;
                messages.put(key, m);
            }
        } catch( Exception e ) {
            err( "Error reading messages file", e );
        }
    }

    static {
        loadMessages();
        // verbose level
        verboseLevel = PropertyManager.getIntProperty( "com.iv.flash.verbose", 0 );
    }

}
