/*
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.parser;

import java.io.*;
import java.util.*;

import com.iv.flash.api.*;
import com.iv.flash.util.*;

public final class Parser extends FlashBuffer {

    private int tagStartPos;
    private int tagDataPos;
    private int tagLength;
    private int tagEndPos;
    private int tagCode;

    // file
    private FlashFile file;

    public Parser() {
    }

    public boolean getBool() {
        return getBits(1) == 1;
    }

    public int getTag() {
        tagStartPos = getPos();
        int code = getUWord();
        int length = code & 0x3f;
        code = code >> 6;

        if( length == 0x3f ) {
            length = getUDWord();
        }

        tagDataPos = getPos();
        tagLength  = length;
        tagCode    = code;
        tagEndPos  = tagDataPos+length;
        return code;
    }

    public void addDef( FlashDef def ) {
        file.addDef(def);
    }

    public FlashDef getDef( int id ) {
        return file.getDef(id);
    }

    public void addDefToLibrary( String name, FlashDef def ) {
        file.addDefToLibrary( name, def );
    }

    public FlashDef getDefFromLibrary( String name ) {
        return file.getDefFromLibrary( name );
    }

    public int getTagStartPos() {
        return tagStartPos;
    }
    public int getTagDataPos() {
        return tagDataPos;
    }
    public int getTagLength() {
        return tagLength;
    }
    public int getTagEndPos() {
        return tagEndPos;
    }
    public int getTagCode() {
        return tagCode;
    }

    public FlashObject newUnknownTag() {
        return new UnparsedTag( tagCode, getBuf(), tagStartPos, tagEndPos );
    }

    public void skipLastTag() {
         setPos( tagEndPos );
    }

    public FlashFile getFile() {
        return file;
    }

    /**
     * Parse input stream
     */
    public void parseStream( InputStream is, FlashFile file ) throws IVException, IOException {

        byte[] fileHdr = new byte[8];

        if( is.read(fileHdr, 0, 8) != 8 ) {
            throw new IVException( "cantReadHeader", file.getFullName() );
        }

        if( fileHdr[0] != 'F' || fileHdr[1] != 'W' || fileHdr[2] != 'S' ) {
            throw new IVException( "illegalHeader", file.getFullName() );
        }

        // get the file size.
        int fileSize = Util.getUDWord(fileHdr[4], fileHdr[5], fileHdr[6], fileHdr[7]);

        if( fileSize < 21 ) {
            throw new IVException( "fileTooShort", file.getFullName() );
        }

        byte[] fileBuf = new byte[fileSize];

        // Copy the data already read from the file.
        System.arraycopy(fileHdr, 0, fileBuf, 0, 8);

        // Read the rest of the file into the buffer.
        int cur = 8;
        int len = 0;
        int left = fileSize - 8;
        while (left > 0 && (len = is.read(fileBuf, cur, left)) != -1) {
            cur += len;
            left -= len;
        }
        if (cur != fileSize)
            throw new IVException( "fileTooShort", file.getFullName() );
        is.close();

        parseBuffer( fileBuf, file );
    }

    /**
     * Parse FlashBuffer
     */
    public void parseBuffer( FlashBuffer fb, FlashFile file ) throws IVException {
        this.file = file;
        init( fb.getBuf(), 0, fb.getSize() );
        parse();
    }

    /**
     * Parse array
     */
    public void parseBuffer( byte[] buffer, FlashFile file ) throws IVException {
        this.file = file;
        init( buffer, 0, buffer.length );
        parse();
    }

    protected void parse() throws IVException {

        // get version
        file.setVersion( (int) getByteAt(3) );

        setPos(8);

        // get frame size
        file.setFrameSize( Rect.parse(this) );

        // get frame rate
        file.setFrameRate( getUWord() );

        // parse main script
        file.setMainScript( Script.parse( this, true ) );
    }

}
