/*
 * $Id: IVUrl.java,v 1.6 2001/07/25 22:44:45 skavish Exp $
 *
 * ==========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.url;

import java.io.*;
import java.net.*;
import java.util.*;
import com.iv.flash.api.*;
import com.iv.flash.util.*;

/**
 * Abstract generator url
 * <p>
 * Usage example:<br>
 * <pre><code>
 * // creates url to file 'data.txt' in the same directory where 'file' is
 * IVUrl url = IVUrl.newUrl( "data.txt", file );
 *
 * // creates url to http://server/getdata.cgi
 * IVUrl url = IVUrl.newUrl( "http://server/getdata.cgi?parm=text" );
 *
 * </code></pre>
 */
public abstract class IVUrl {

    /**
     * Creates new url from specified string
     *
     * @param surl   absolute url
     * @return generator url
     * @exception IVException
     */
    public static IVUrl newUrl( String surl ) throws IVException {
        return newUrl( surl, null );
    }

    /**
     * Creates new url relative from specified file
     *
     * @param surl   relative or absolute url
     * @return generator url
     * @exception IVException
     */
    public static IVUrl newUrl( String surl, FlashFile flashFile ) throws IVException {
        if( surl.startsWith("fgjdbc:///") ) {
            return new JDBCUrl( surl );
        } else if( surl.startsWith("fgjava:///") ) {
            return new JavaUrl( surl );
        } else if( surl.startsWith("fgftp:///") ) {
            return null;//new FTPURL( surl );
        }
        else {
            try {
                URL url = new URL(surl);
                return new URLUrl( url );
            } catch( MalformedURLException e ) {
                return new FileUrl( surl, flashFile );
            }
        }
    }

    /**
     * Returns name of the url.
     *
     * @return name of this url
     */
    public abstract String getName();

    /**
     * Returns input stream for this url
     *
     * @return url input stream
     * @exception IOException
     */
    public abstract InputStream getInputStream() throws IOException;

    /**
     * Returns parameter of the url by name
     * <p>
     * Parameters are standard url parameters like 'parm' in: http://server/getdata.cgi?parm=text
     *
     * @param name   parameter name
     * @return parameter value or null
     */
    public String getParameter( String name ) {
        return null;
    }

    /**
     * Returns timestamp of last time this url was modified
     *
     * @return last modified time
     */
    public long lastModified() {
        return 0;
    }

    /**
     * Refreshes this url. Updates last modified.
     */
    public void refresh() {
    }

    /**
     * Returns true if this url can provide data in name,value format right away
     * without parsing input stream. If returns true must implement {@link #getData} too.
     * @see #getData
     */
    public boolean hasDataReady() {
        return false;
    }

    /**
     * Returns data in name,value format ready to be used.
     * @see #hasDataReady
     */
    public String[][] getData() throws IOException {
        return null;
    }

    public String toString() {
        return getName();
    }

    /**
     * Returns url reference (the part of url after #)
     *
     * @return reference
     */
    public String getRef() {
        return null;
    }

    /**
     * Decodes specified url and parses its parameters
     *
     * @param s      url
     * @return hashtable of parameters
     */
    protected Hashtable parseQueryString( String s ) {
        return parse( decode( s ) );
    }

    /**
     * Parses parameters of url-string
     * <p>
     * Parameters are pairs: name=value, separated by ampersand
     *
     * @param s      url
     * @return hashtable of parameters
     */
    protected Hashtable parse( String s ) {
        int idx = s.indexOf('?');
        if( idx == -1 ) return null;
        s = s.substring( idx+1 );
        Hashtable table = new Hashtable();
        StringTokenizer st = new StringTokenizer(s,"&");
        while( st.hasMoreTokens() ) {
            String token = st.nextToken();
            idx = token.indexOf('=');
            if( idx == -1 ) return null;
            table.put( token.substring(0, idx), token.substring(idx+1) );
        }
        return table;
    }

    /**
     * Decodes url-encoded string
     *
     * @param s      url-encoded string
     * @return decoded string
     */
    protected String decode( String s ) {
        StringBuffer sb = new StringBuffer(s.length());
        for( int i=0; i<s.length(); i++ ) {
            char ch = s.charAt(i);
            if( ch == '+' ) {
                sb.append(' ');
            } else if( ch == '%' ) {
                ch = (char) Integer.parseInt(s.substring(i+1,i+3), 16);
                sb.append(ch);
                i+=2;
            } else {
                sb.append(ch);
            }
        }
        return new String(sb);
    }

}
