/*
 * $Id: PropertyManager.java,v 1.10 2001/08/02 22:26:34 skavish Exp $
 *
 * ===========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.util;

import com.iv.flash.api.*;
import com.iv.flash.log.*;
import com.iv.flash.cache.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * Property manager.
 * <P>
 * Reads and writes properties from iv.property file.
 * Caches most used properties as static variables
 *
 * @author Dmitry Skavish
 */
public class PropertyManager {

    /**
     * Property com.iv.flash.varCaseSensitive
     */
    public static boolean varCaseSensitive = true;
    /**
     * Property com.iv.flash.symCaseSensitive
     */
    public static boolean symCaseSensitive = false;
    /**
     * Property com.iv.flash.showErrorsInline
     */
    public static boolean showErrorsInline = true;
    /**
     * Property com.iv.flash.textBoundsMMStyle
     */
    public static boolean textMMStyle      = false;
    /**
     * Property com.iv.flash.fontPath
     */
    public static String fontPath          = "fonts/";

    private static Properties myProperties;

    private static void loadCacheProperties( CacheSettings cs, String prefix ) {
        cs.setMaxSize( getIntProperty( prefix+"CacheMaxSize", 0 ) );
        cs.setDefaultExpire( (long) (getDoubleProperty( prefix+"CacheDefaultExpire", 0.0 )*1000) );
        cs.setForce( getBoolProperty( prefix+"CacheForce", false ) );
        cs.setRecycle( getBoolProperty( prefix+"CacheRecycle", false ) );
        cs.setCheckModifiedSince( getBoolProperty( prefix+"CacheCheckModifiedSince", false ) );
    }

    private static void saveCacheProperties( CacheSettings cs, String prefix ) {
        setProperty( prefix+"CacheMaxSize", cs.getMaxSize() );
        setProperty( prefix+"CacheDefaultExpire", cs.getDefaultExpire()/1000.0 );
        setProperty( prefix+"CacheForce", cs.isForce() );
        setProperty( prefix+"CacheRecycle", cs.isRecycle() );
        setProperty( prefix+"CacheCheckModifiedSince", cs.isCheckModifiedSince() );
    }

    /**
     * Initalizes manager.
     * <P>
     * Reads and caches all the properties
     */
    public static void init() {
        load();
        varCaseSensitive = getBoolProperty( "com.iv.flash.varCaseSensitive", true );
        symCaseSensitive = getBoolProperty( "com.iv.flash.symCaseSensitive", false );
        showErrorsInline = getBoolProperty( "com.iv.flash.showErrorsInline", true );
        textMMStyle      = getBoolProperty( "com.iv.flash.textBoundsMMStyle", false );
        fontPath         = getProperty( "com.iv.flash.fontPath", "fonts/" );
    }

    /**
     * Loads iv.properties file and caches "cache" related properties.
     */
    public static void load() {
        try {
            myProperties = new Properties();
            myProperties.load( new FileInputStream( Util.getSysFile( "iv.properties" ) ) );
        } catch( IOException e ) {
            Log.err( "Cannot load properties file", e );
        }
        loadCacheProperties( getRequestCacheSettings(), "com.iv.flash.request" );
        loadCacheProperties( getFontCacheSettings(), "com.iv.flash.font" );
        loadCacheProperties( getMediaCacheSettings(), "com.iv.flash.media" );
        loadCacheProperties( getXMLCacheSettings(), "com.iv.flash.xml" );
    }

    /**
     * Saves properties to iv.properties file.
     *
     * @exception IOException
     */
    public static void save() throws IOException {
        saveCacheProperties( getRequestCacheSettings(), "com.iv.flash.request" );
        saveCacheProperties( getFontCacheSettings(), "com.iv.flash.font" );
        saveCacheProperties( getMediaCacheSettings(), "com.iv.flash.media" );
        saveCacheProperties( getXMLCacheSettings(), "com.iv.flash.xml" );
        setProperty( "com.iv.flash.varCaseSensitive", varCaseSensitive );
        setProperty( "com.iv.flash.symCaseSensitive", symCaseSensitive );
        setProperty( "com.iv.flash.showErrorsInline", showErrorsInline );
        setProperty( "com.iv.flash.textBoundsMMStyle", textMMStyle );
        setProperty( "com.iv.flash.fontPath", fontPath );
        // has to be replaced with store
        myProperties.save( new FileOutputStream( Util.getSysFile( "iv.properties" ) ), "" );
    }

    /**
     * Sets property by name.
     *
     * @param name   name of the property
     * @param value  value of the property
     */
    public static void setProperty( String name, String value ) {
        myProperties.put(name, value);
    }

    /**
     * Sets integer property by name.
     *
     * @param name   name of the property
     * @param value  value of the property
     */
    public static void setProperty( String name, int value ) {
        myProperties.put(name, new Integer(value).toString());
    }

    /**
     * Sets boolean property by name.
     *
     * @param name   name of the property
     * @param value  value of the property
     */
    public static void setProperty( String name, boolean value ) {
        myProperties.put(name, new Boolean(value).toString());
    }

    /**
     * Sets double property by name.
     *
     * @param name   name of the property
     * @param value  value of the property
     */
    public static void setProperty( String name, double value ) {
        myProperties.put(name, new Double(value).toString());
    }

    /**
     * Gets property by name.
     *
     * @param name   name of the property
     * @param def    default value to be returned if there is no such property
     * @return value of the property or provided default value
     */
    public static String getProperty( String name, String def ) {
        String retVal = null;
        
        try{
            retVal = myProperties.getProperty(name, def);
        }
        catch( Exception e ){
            retVal = "";
        }
        
        return (retVal == null) ? "" : retVal;
    }

    /**
     * Gets property by name.
     *
     * @param name   name of the property
     * @return value of the property or null
     */
    public static String getProperty( String name ) {
        String retVal = null;
        
        try{
            retVal = myProperties.getProperty(name);
        }
        catch( Exception e ){
            retVal = "";
        }
        
        //return myProperties.getProperty(name);
        
        return (retVal == null) ? "" : retVal;
    }

    /**
     * Gets integer property by name.
     *
     * @param name   name of the property
     * @param def    default value to be returned if there is no such property
     * @return value of the property or provided default value
     */
    public static int getIntProperty( String property, int def ) {
        return Util.toInt( getProperty(property), def );
    }

    /**
     * Gets long property by name.
     *
     * @param name   name of the property
     * @param def    default value to be returned if there is no such property
     * @return value of the property or provided default value
     */
    public static long getLongProperty( String property, long def ) {
        return Util.toLong( getProperty(property), def );
    }

    /**
     * Gets boolean property by name.
     *
     * @param name   name of the property
     * @param def    default value to be returned if there is no such property
     * @return value of the property or provided default value
     */
    public static boolean getBoolProperty( String property, boolean def ) {
        return Util.toBool( getProperty(property), def );
    }

    /**
     * Gets double property by name.
     *
     * @param name   name of the property
     * @param def    default value to be returned if there is no such property
     * @return value of the property or provided default value
     */
    public static double getDoubleProperty( String property, double def ) {
        return Util.toDouble( getProperty(property), def );
    }

    /**
     * Returns request cache settings.
     *
     * @return request cache settings
     */
    public static CacheSettings getRequestCacheSettings() {
        return RequestCache.getSettings();
    }

    /**
     * Returns font cache settings.
     *
     * @return font cache settings
     */
    public static CacheSettings getFontCacheSettings() {
        return FontCache.getSettings();
    }

    /**
     * Returns media cache settings.
     *
     * @return media cache settings
     */
    public static CacheSettings getMediaCacheSettings() {
        return MediaCache.getSettings();
    }

    /**
     * Returns xml cache settings.
     *
     * @return xml cache settings
     */
    public static CacheSettings getXMLCacheSettings() {
        return XMLCache.getSettings();
    }

//------------------------------------------------------------------------------
    public static void main( String args[] ){
        PropertyManager m = null;
        
        m.init();
        
    }
}
