/*
 * $Id: TextLayout.java,v 1.10 2001/08/02 22:26:34 skavish Exp $
 *
 * ===========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.util;

import com.iv.flash.api.*;
import com.iv.flash.api.text.*;
import java.io.*;

/**
 * This class does simple text layout in a box.
 * <P>
 *
 * @author Dmitry Skavish
 * @see com.iv.flash.api.text.Text
 * @see com.iv.flash.api.text.Font
 * @see com.iv.flash.api.text.FontDef
 * @see com.iv.flash.api.text.TextRecord
 * @see com.iv.flash.api.text.TextStyleChangeRecord
 */
public final class TextLayout {

    /**
     * One line of text.
     * <p>
     * Contains several style runs (text of different font and style)
     */
    public final class TextLine extends FlashItem {
        public IVVector records = new IVVector(4);
        private TextItem lastItem;
        private TextRecord lastRecord;

        public TextLine() {}

        /**
         * Creates new text record.
         */
        public void newRecord() {
            if( lastItem != null && lastRecord != null && lastRecord.getSize() != 0 ) {
                records.addElement( getStyleChange( lastItem ) );
            }
            if( lastRecord != null && lastRecord.getSize() != 0 ) {
                records.addElement( lastRecord );
            }
            lastItem = curItem;
            lastRecord = new TextRecord( 128 );
        }

        /**
         * Creates text style change record from specified text item
         * and sets all the properties, like font, color and height as
         * current properties.
         *
         * @param item   item for which to create style change record
         * @return new style change record
         */
        private TextStyleChangeRecord getStyleChange( TextItem item ) {
            // create text style change record
            TextStyleChangeRecord ts = new TextStyleChangeRecord();

            if( item.font != lastFont || item.height != lastHeight ) {
                ts.setFont( item.font );
                ts.setHeight( item.height );
            }
            if( !item.color.equals(lastColor) ) {
                ts.setColor( item.color );
            }

            lastFont = item.font;
            lastColor = item.color;
            lastHeight = item.height;

            return ts;
        }

        /**
         * Adds new character to last text record.
         * <P>
         * If last text record is overflowed then creates new text record
         *
         * @param ch     character to be added
         * @param index  index of character in the font
         * @param adv    advance value of the character
         */
        public void add( char ch, int index, int adv ) {
            lastRecord.add( ch, index, adv );
            if( lastRecord.getSize() >= 127 ) {
                newRecord();
                lastItem = null;
            }
        }

        /**
         * Mark current position in current text record.
         * <P>
         * Used for rolling back if word does not fit in the text rectangle.
         */
        public void markPosition() {
            markedPosition = lastRecord.getSize();
        }

        /**
         * Rollbacks current position to the marked one.
         */
        public void rollBack() {
            lastRecord.setSize( markedPosition );
        }

        /**
         * Does end-of-line.
         * <P>
         * Flushes current text record
         *
         * @param x      x position of this line
         * @param y      y position of this line
         */
        public void endLine( int x, int y ) {
            newRecord();    // just flush temp records
            if( records.size() > 0 ) {
                TextStyleChangeRecord ts = (TextStyleChangeRecord) records.elementAt(0);
                ts.setX(x);
                ts.setY(y);
            }
        }

        /**
         * Returns maximum advance value from all the records.
         *
         * @return maximum advance value from all the records
         */
        public int getMaxAdvance() {
            int max = Integer.MIN_VALUE;
            for( int i=0; i<records.size(); i++ ) {
                Object o = records.elementAt(i);
                if( o instanceof TextRecord ) {
                    TextRecord tr = (TextRecord) o;
                    int adv = tr.getMaxAdvance();
                    if( adv > max ) max = adv;
                }
            }
            return max;
        }

        /**
         * Returns maximum character index from all the records.
         *
         * @return maximum character index from all the records
         */
        public int getMaxIndex() {
            int max = Integer.MIN_VALUE;
            for( int i=0; i<records.size(); i++ ) {
                Object o = records.elementAt(i);
                if( o instanceof TextRecord ) {
                    TextRecord tr = (TextRecord) o;
                    int idx = tr.getMaxIndex();
                    if( idx > max ) max = idx;
                }
            }
            return max;
        }

        /**
         * Returns width of this line.
         *
         * @return width if this line in twixels
         */
        public int getWidth() {
            if( records.size() == 0 ) return 0;
            TextStyleChangeRecord ts = (TextStyleChangeRecord) records.elementAt(0);
            int width = ts.getX();
            for( int i=1; i<records.size(); i++ ) {
                Object o = records.elementAt(i);
                if( o instanceof TextRecord ) {
                    width += ((TextRecord)o).getWidth();
                }
            }
            return width;
        }

        public void printContent( PrintStream out, String indent ) {
            out.println( indent+"TextLine: size="+records.size()+" x="+x+" y="+y );
            records.printContent(out, indent+"   ");
        }

        public void write( FlashOutput fob ) {
            records.write(fob);
        }

        protected FlashItem copyInto( FlashItem item, ScriptCopier copier ) {
            ((TextLine)item).records = records.getCopy(copier);
            // there is no need to copy lastItem and lastRecord (?)
            return item;
        }

        public FlashItem getCopy( ScriptCopier copier ) {
            return copyInto( new TextLine(), copier );
        }
    }

    private int markedPosition = 0; // used only by TextLine

    private TextItem curItem;
    private TextLine curLine;
    private int line_width;
    private int line_window;
    private int x;
    private int y;
    private int max_ascent;
    private int max_descent;
    private int max_linesp;

    private Font lastFont;
    private Color lastColor;
    private int lastHeight = -1;
    private int rect_width;

    private Text myText;
    private Rect bounds;
    private IVVector lines = new IVVector();

    private void reCalcHeights() {
        Font font = curItem.font;
        int height = curItem.height;
        max_ascent = (font.ascent * height) / 1024;
        max_linesp = curItem.linesp;
        max_descent = (font.descent * height) / 1024 + max_linesp;
    }

    private void endLine() {
        if( curLine != null ) {
            y += max_ascent;    // advance y to baseline
            curLine.endLine(x, y);
            y += max_descent;   // advance y to next line
            curLine = null;
        }
    }

    private void newLine() {
        endLine();
        reCalcHeights();
        curLine = new TextLine();
        lines.addElement( curLine );
        curLine.newRecord();
        line_width = 0;
        x = curItem.marginleft;
        line_window = rect_width-(curItem.marginleft+curItem.marginright);
    }

    private void newParagraph() {
        newLine();
        line_window -= curItem.indent;
        x += curItem.indent;
    }

    /**
     * Creates new text layout.
     *
     * @param myText text to be layed out
     * @param bounds rectangle to be used for laying the text out in
     */
    public TextLayout( Text myText, Rect bounds ) {
        this.myText = myText;
        this.bounds = bounds;
    }

    /**
     * Retrieves vector of text records layed out by this
     * text layout for the specified font.
     *
     * @param font   font of the records to be retrieved
     * @return vector of {@link com.iv.flash.api.text.TextRecord}s
     */
    public IVVector getTextRecords( Font font ) {
        IVVector trs = new IVVector();
        Font lastFont = null;
        for( int i=0; i<lines.size(); i++ ) {
            TextLine line = (TextLine) lines.elementAt(i);
            for( int k=0; k<line.records.size(); k++ ) {
                Object o = line.records.elementAt(k);
                if( o instanceof TextStyleChangeRecord ) {
                    Font f = ((TextStyleChangeRecord)o).getFont();
                    if( f != null ) lastFont = f;
                } else {
                    if( lastFont == font ) trs.addElement(o);
                }
            }
        }
        return trs;
    }

    private void alignLine() {
        switch( curItem.align ) {
            case 0: // left
                break;
            case 1: // right
                x += line_window-line_width;
                break;
            case 2: // center
                x += (line_window-line_width)/2;
                break;
            case 3: // justify
                // ........
                break;
        }
    }

    /**
     * Does text layout.
     */
    public void layout() {
        IVVector items = myText.getTextItems();
//        System.out.println( "layout of '"+((TextItem)items.elementAt(0)).text+"'" );

        rect_width = bounds.getWidth();

        y = 0;

        curItem = (TextItem) items.elementAt(0);

        newParagraph();

        int i=0;
        for(;;) {
            Font font = curItem.font;
            int height = curItem.height;
            int ascent = (font.ascent * height) / 1024;
            int descent = (font.descent * height) / 1024 + curItem.linesp;
            if( ascent > max_ascent ) max_ascent = ascent;
            if( descent > max_descent ) {
                max_descent = descent;
                max_linesp = curItem.linesp;
            }

            String text = curItem.text;
            boolean isNowWord = false;
            int word_width = 0;
            int start_word = 0;

            int text_len = text.length();
            for( int k=0; k<text_len; k++ ) {
                char ch = text.charAt(k);
                boolean isWord = isWord(ch);
                if( !isNowWord && isWord  ) { // start new word
                    curLine.markPosition();
                    word_width = 0;
                    start_word = k;
                    isNowWord = true;
                }
                isNowWord = isWord;
                if( ch == '\r' || ch == '\n' ) {
                    alignLine();
                    if( k == text_len-1 ) endLine();
                    else {
                        char ch1 = text.charAt(k+1);
                        if( ch != ch1 && (ch1 == '\r' || ch1 == '\n') ) k++;
                        if( k == text_len-1 ) endLine();
                        else newParagraph();
                    }
                } else {
                    int idx = font.getIndex(ch);
                    int ch_adv = font.getAdvanceValue(idx);
                    if( k != text_len-1 ) ch_adv += font.getKerning(ch, text.charAt(k+1));
                    int adv = (ch_adv * height) / 1024;
                    adv += curItem.kerning;
                    line_width += adv;
                    word_width += adv;
                    if( line_width <= line_window ) {
                        curLine.add(ch,idx,adv);
                    } else {
                        if( isNowWord ) {
                            if( word_width > line_window ) {
                                // split anyway, because the word does not fit the window at all
                                line_width -= adv;
                                k--;
                                newLine();
                            } else {
                                line_width -= word_width;
                                curLine.rollBack();
                                k = start_word-1;
                                newLine();
                                isNowWord = false;
                            }
                        } else {
                            line_width -= adv;
                            k--;
                            newLine();
                        }
                    }
                }
            }

            i++;
            if( i >= items.size() ) break;
            curItem = (TextItem) items.elementAt(i);
            if( curLine == null ) newParagraph();
            else curLine.newRecord();
        }
        alignLine();
        endLine();

        y -= max_linesp;

        // calculate bounds
        int maxX = Integer.MIN_VALUE;
        for( int l=0; l<lines.size(); l++ ) {
            TextLine line = (TextLine) lines.elementAt(l);
            int max = line.getWidth();
            if( max > maxX ) maxX = max;
        }

        // check what kind of behavior is expected: MMGen or JGen and
        //        create appropriate bounds
        if( PropertyManager.textMMStyle ) {
            // set bounds to be exactly equal to the text (MMGen style)
            bounds.setXMax( bounds.getXMin()+maxX );
            bounds.setYMax( bounds.getYMin()+y );
        } else {
            // set bounds to be equal whatever came from the template (JGen style)
            if( maxX > bounds.getWidth() )
                bounds.setXMax( bounds.getXMin()+maxX );
            if( y > bounds.getHeight() )
                bounds.setYMax( bounds.getYMin()+y );
        }

        myText.setBounds(bounds);
    }

    protected int getNGlyphBits() {
        int maxIdx = 0;
        for( int l=0; l<lines.size(); l++ ) {
            TextLine line = (TextLine) lines.elementAt(l);
            int max = line.getMaxIndex();
            if( max > maxIdx ) maxIdx = max;
        }

        return Util.getMinBitsU(maxIdx);
    }

    protected int getNAdvanceBits() {
        int maxAdv = Integer.MIN_VALUE;
        for( int l=0; l<lines.size(); l++ ) {
            TextLine line = (TextLine) lines.elementAt(l);
            int max = line.getMaxAdvance();
            if( max > maxAdv ) maxAdv = max;
        }

        return Util.getMinBitsS(maxAdv);
    }

    public void write( FlashOutput fob ) {
        int nGlyphBits = getNGlyphBits();
        int nAdvanceBits = getNAdvanceBits();
        fob.writeByte(nGlyphBits);
        fob.writeByte(nAdvanceBits);
        fob.setUserData( new int[] {nGlyphBits, nAdvanceBits} );

        for( int i=0; i<lines.size(); i++ ) {
            TextLine line = (TextLine) lines.elementAt(i);
            line.write(fob);
        }
        fob.writeByte(0);
    }

    public TextLayout getCopy( ScriptCopier copier ) {
        TextLayout tl = new TextLayout(myText,bounds);
        tl.lines = lines.getCopy(copier);
        return tl;
    }

    private static boolean isWord( char ch ) {
        return Character.isLetterOrDigit(ch);
    }
}
