/*
 * $Id: XMLContext.java,v 1.7 2001/08/02 22:26:34 skavish Exp $
 *
 * ===========================================================================
 *
 * The JGenerator Software License, Version 1.0
 *
 * Copyright (c) 2000 Dmitry Skavish (skavish@usa.net). All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *    "This product includes software developed by Dmitry Skavish
 *     (skavish@usa.net, http://www.flashgap.com/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The name "The JGenerator" must not be used to endorse or promote
 *    products derived from this software without prior written permission.
 *    For written permission, please contact skavish@usa.net.
 *
 * 5. Products derived from this software may not be called "The JGenerator"
 *    nor may "The JGenerator" appear in their names without prior written
 *    permission of Dmitry Skavish.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL DMITRY SKAVISH OR THE OTHER
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

package com.iv.flash.util;

import com.iv.flash.api.*;
import com.iv.flash.log.*;

import java.util.*;

import javax.xml.transform.TransformerException;

import org.w3c.dom.*;
import org.w3c.dom.traversal.NodeIterator;

import org.apache.xpath.*;
import org.apache.xpath.objects.XObject;

/**
 * XML context.
 * <P>
 * Represents one xml node
 *
 * @author Dmitry Skavish
 * @see StandardContext
 */
public class XMLContext extends Context {

    private Node node;
    private XPathContext xpathContext;

    public XMLContext() {}

    /**
     * Creates xml context with specified parent and xml node.
     *
     * @param parent parent context
     * @param node   xml node represented by this context
     */
    public XMLContext( Context parent, Node node ) {
        setParent( parent );
        this.node = node;
        this.xpathContext = new XPathContext();
    }

    /**
     * Gets value of given xpath expression.
     * <P>
     * Tries to resolve given string in this context as XPath expression,
     * otherwise passes it to its parent.
     *
     * @param name   XPath expression or variable name (in this last case it will be passed to parent)
     * @return string representation of result of xpath execution or result from the parent or null
     */
    public String _getValue( String name ) {
        try {
            XPath xpath = XMLHelper.getXPath( name );
            XObject xo = XMLHelper.evalXPath(xpathContext, node, xpath);
            if( !isUndefined(xo) ) return XMLHelper.getXObjectData(xo);
        } catch( TransformerException e ) {
            // ignore this exception, it usually means that it is not an xpath, but just a variable
        } catch( Exception e ) {
            Log.log(e);
        }
        Context parent = getParent();
        if( parent != null ) return parent._getValue( name );
        return null;
    }

    /**
     * Executes specified XPath in this context or nearest xml parent.
     *
     * @param xpath compiled XPath expression
     * @return string representation of result of xpath execution or empty string
     */
    public String getValue( XPath xpath ) {
        try {
            XObject xo = eval(xpath);
            return XMLHelper.getXObjectData(xo);
        } catch( TransformerException e ) {
            // ignore this exception, it usually means that it is not an xpath, but just a variable
            return null;
        } catch( Exception e ) {
            Log.log(e);
            return null;
        }
    }

    /**
     * Returns nearest xml parent skipping all non-xml.
     *
     * @return nearest xml parent context
     */
    public XMLContext getXMLParent() {
        Context parent = getParent();
        while( parent != null && !(parent instanceof XMLContext) )
            parent = parent.getParent();
        return (XMLContext) parent;
    }

    /**
     * Returns context node.
     *
     * @return context node
     */
    public Node getContextNode() {
        return node;
    }

    /**
     * Compiles given string into XPath and execute it in this context or its parent.
     *
     * @param expr   supposedly XPath expression
     * @return result of XPath execution
     * @exception TransformerException
     */
    public XObject eval( String expr ) throws TransformerException {
        XPath xpath = XMLHelper.getXPath( expr );
        return eval( xpath );
    }

    /**
     * Executes specified XPath in this context or nearest xml parent.
     *
     * @param xpath  compiled XPath expression
     * @return result of xpath execution
     * @exception TransformerException
     */
    public XObject eval( XPath xpath ) throws TransformerException {
        //System.out.println( node.getNodeName()+".eval('"+xpath.getExpression()+"')" );
        XObject xo = XMLHelper.evalXPath(xpathContext, node, xpath);
        //System.out.println( "value: '"+xo.toString()+"'" );
        if( !isUndefined(xo) ) return xo;
        XMLContext parent = getXMLParent();
        if( parent != null ) return parent.eval( xpath );
        return xo;
    }

    /**
     * Check whether the specified XObject undefined or not.
     *
     * @param xo     XObject, usually results of XPath execution
     * @return true if specified xobject is undefined
     * @exception TransformerException
     */
    public boolean isUndefined( XObject xo ) throws TransformerException {
        int t = xo.getType();
        return t == XObject.CLASS_UNKNOWN ||
               (t == XObject.CLASS_NODESET && XMLHelper.getNodeList(xo).getLength() == 0);
    }

    /**
     * Evaluates XPath in this context to list of nodes.
     *
     * @param expr   string considered to be XPath expression
     * @return result of xpath execution - list of nodes
     * @exception TransformerException
     */
    public NodeList selectNodeList( String expr ) throws TransformerException {
        return XMLHelper.getNodeList( eval(expr) );
    }

    /**
     * Evaluates XPath in this context to list of nodes.
     *
     * @param xpath     XPath expression
     * @return result of xpath execution - list of nodes
     * @exception TransformerException
     */
    public NodeList selectNodeList( XPath xpath ) throws TransformerException {
        return XMLHelper.getNodeList( eval(xpath) );
    }

    /**
     * Evaluates XPath in this context to one node.
     *
     * @param expr     string considered to be XPath expression
     * @return xml node
     * @exception TransformerException
     */
    public Node selectSingleNode( String expr ) throws TransformerException {
        NodeList nl = selectNodeList(expr);
        if( nl.getLength() > 0 ) return nl.item(0);
        return null;
    }

    /**
     * Evaluates XPath in this context to one node.
     *
     * @param xpath     XPath expression
     * @return xml node
     * @exception TransformerException
     */
    public Node selectSingleNode( XPath xpath ) throws TransformerException {
        NodeList nl = selectNodeList(xpath);
        if( nl.getLength() > 0 ) return nl.item(0);
        return null;
    }

}
